"""

This module contains basic set-up stuff (making directories, parsing config etc.) used by all the scripts 
in bin/ (nemo, nemoMass, nemoSelFn etc.).

"""

import os
import sys
import yaml
import copy
import astropy.io.fits as pyfits
import astropy.table as atpy
from astLib import astWCS, astImages
from nemo import signals
import numpy as np
import pickle
import time
from . import maps

#------------------------------------------------------------------------------------------------------------
def parseConfigFile(parDictFileName, verbose = False):
    """Parse a Nemo .yml config file.
    
    Args:
        parDictFileName (:obj:`str`): Path to a nemo .yml configuration file.
        verbose (:obj:`bool`, optional): If True, warning messages may be printed to the console, if there
            are any.
    
    Returns:
        A dictionary of parameters.
    
    """
    
    if verbose:
        print(">>> Parsing config file %s" % (parDictFileName))
    with open(parDictFileName, "r") as stream:
        parDict=yaml.safe_load(stream)
        # We've moved masks out of the individual map definitions in the config file
        # (makes config files simpler as we would never have different masks across maps)
        # To save re-jigging how masks are treated inside filter code, add them back to map definitions here
        maskKeys=['pointSourceMask', 'surveyMask', 'flagMask', 'maskPointSourcesFromCatalog', 'apodizeUsingSurveyMask',
                  'maskSubtractedPointSources', 'RADecSection', 'maskHoleDilationFactor', 'reprojectToTan']
        for mapDict in parDict['unfilteredMaps']:
            for k in maskKeys:
                if k in parDict.keys():
                    mapDict[k]=parDict[k]
                else:
                    mapDict[k]=None
            # Also add key for type of weight map (inverse variance is default for enki maps)
            if 'weightsType' not in mapDict.keys():
                mapDict['weightsType']='invVar'
        # Apply global filter options (defined in allFilters) to mapFilters
        # Note that anything defined in mapFilters has priority
        # Bit ugly... we only support up to three levels of nested dictionaries...
        if 'allFilters' in parDict.keys():
            mapFiltersList=[]
            for filterDict in parDict['mapFilters']:
                newDict=copy.deepcopy(parDict['allFilters'])
                for key in filterDict.keys():
                    if type(filterDict[key]) == dict: 
                        if key not in newDict.keys():
                            newDict[key]={}
                        for subkey in filterDict[key].keys():
                            if type(filterDict[key][subkey]) == dict:
                                if subkey not in filterDict[key].keys():
                                    newDict[key][subkey]={}
                                for subsubkey in filterDict[key][subkey].keys():
                                    if type(filterDict[key][subkey][subsubkey]) == dict:
                                        if subsubkey not in filterDict[key][subkey].keys():
                                            newDict[key][subkey][subsubkey]={}                                    
                                    # No more levels please...
                                    newDict[key][subkey][subsubkey]=filterDict[key][subkey][subsubkey]                                    
                            else:
                                newDict[key][subkey]=filterDict[key][subkey]
                    else:
                        newDict[key]=filterDict[key]
                mapFiltersList.append(newDict)
            parDict['mapFilters']=mapFiltersList
        # We always need RMSMap and freqWeightsMap to do any photometry
        # So we may as well force inclusion if they have not been explicitly given
        if 'photFilter' not in parDict.keys():
            # This is to allow source finding folks to skip this option in .yml
            # (and avoid having 'fixed_' keywords in output (they have only one filter scale)
            parDict['photFilter']=None
        else:
            photFilter=parDict['photFilter']
            for filtDict in parDict['mapFilters']:
                if filtDict['label'] == photFilter:
                    filtDict['params']['saveRMSMap']=True
                    filtDict['params']['saveFreqWeightMap']=True
                    filtDict['params']['saveFilter']=True
        # Global noise mask catalog, if given, now goes into filter settings where it belongs
        if 'noiseMaskCatalog' in parDict.keys() and parDict['noiseMaskCatalog'] is not None:
            for filtDict in parDict['mapFilters']:
                filtDict['params']['noiseMaskCatalog']=parDict['noiseMaskCatalog']
        # tileNames must be case insensitive in .yml file 
        # we force upper case here (because FITS will anyway)
        if 'tileDefinitions' in parDict.keys():
            if type(parDict['tileDefinitions']) == list:
                for tileDef in parDict['tileDefinitions']:
                    tileDef['tileName']=tileDef['tileName'].upper()
        if 'tileNameList' in parDict.keys():
            newList=[]
            for entry in parDict['tileNameList']:
                newList.append(entry.upper())
            parDict['tileNameList']=newList
        if 'reprojectToTan' not in parDict.keys():
            parDict['reprojectToTan']=False
        # We shouldn't have to give this unless we're using it
        if 'catalogCuts' not in parDict.keys():
            parDict['catalogCuts']=[]
        # Don't measure object shapes by default
        if 'measureShapes' not in parDict.keys():
            parDict['measureShapes']=False
        # Don't reject objects in map border areas by default
        if 'rejectBorder' not in parDict.keys():
            parDict['rejectBorder']=0
        # By default, undo the pixel window function
        if 'undoPixelWindow' not in parDict.keys():
            parDict['undoPixelWindow']=True
        if 'fitQ' not in parDict.keys():
            parDict['fitQ']=False
        if 'calcSelFn' not in parDict.keys():
            parDict['calcSelFn']=False
        # We need a better way of giving defaults than this...
        if 'selFnOptions' in parDict.keys() and 'method' not in parDict['selFnOptions'].keys():
            parDict['selFnOptions']['method']='fast'
        if 'selFnOptions' in parDict.keys() and parDict['selFnOptions']['method'] not in ['fast', 'injection']:
                raise Exception("Valid completeness estimation methods are 'fast' or 'injection' - edit selFnOptions['method'] in config.")
        if 'selFnOptions' in parDict.keys() and 'QSource' not in parDict['selFnOptions'].keys():
            if parDict['fitQ'] == True:
                parDict['selFnOptions']['QSource']='fit'
            else:
                parDict['selFnOptions']['QSource']='injection'
        # Check of tile definitions
        if 'useTiling' not in list(parDict.keys()):
            parDict['useTiling']=False
        if 'tileDefinitions' in parDict.keys() and type(parDict['tileDefinitions']) == list:
            checkList=[]
            for entry in parDict['tileDefinitions']:
                if entry['tileName'] in checkList:
                    raise Exception("Duplicate tileName '%s' in tileDefinitions - fix in config file" % (entry['tileName']))
                checkList.append(entry['tileName'])
        if 'stitchTiles' not in list(parDict.keys()):
            if parDict['useTiling'] == True:
                parDict['stitchTiles']=True
            else:
                parDict['stitchTiles']=False
        # Optional override of default GNFW parameters (used by Arnaud model), if used in filters given
        if 'GNFWParams' not in list(parDict.keys()):
            parDict['GNFWParams']='default'
        for filtDict in parDict['mapFilters']:
            filtDict['params']['GNFWParams']=parDict['GNFWParams']
        # Optional forced photometry
        if 'forcedPhotometryCatalog' not in parDict.keys():
            parDict['forcedPhotometryCatalog']=None
        # Used for finding and removing rings around bright sources
        if 'removeRings' not in parDict.keys():
            parDict['removeRings']=True
        if 'ringThresholdSigma' not in parDict.keys():
            parDict['ringThresholdSigma']=3
        # Applies to source injection recover sims only (whether print message or trigger exception)
        if 'haltOnPositionRecoveryProblem' not in parDict.keys():
            parDict['haltOnPositionRecoveryProblem']=False
        # Mass/scaling relation/cosmology options - set fiducial values here if not chosen in config
        # NOTE: We SHOULD use M200c not M500c here (to avoid CCL Tinker08 problem)
        # But we don't, currently, as old runs/tests used M500c and Arnaud-like scaling relation
        if 'massOptions' not in parDict.keys():
            parDict['massOptions']={}
        defaults={'tenToA0': 4.95e-5, 'B0': 0.08, 'Mpivot': 3.0e+14, 'sigma_int': 0.2,
                  'relativisticCorrection': True, 'rhoType': 'critical', 'delta': 500,
                  'H0': 70.0, 'Om0': 0.3, 'Ob0': 0.05, 'sigma8': 0.80, 'ns': 0.95,
                  'concMassRelation': 'Bhattacharya13'}
        for key in defaults:
            if key not in parDict['massOptions'].keys():
                parDict['massOptions'][key]=defaults[key]

    # Stuff which is now mandatory... left here until we update everywhere in code + docs
    parDict['stitchTiles']=True

    # This isn't actually being used, but has been left in for now
    parDict['_file_last_modified_ctime']=os.path.getctime(parDictFileName)
    
    # To aid user friendliness - spot any out-of-date / removed / renamed parameters here
    # Use None for those that are totally removed
    oldKeyMap={'makeTileDir': 'useTiling', 'tileDefLabel': None, 'twoPass': None,
               'clusterInjectionModels': 'sourceInjectionModels'}
    for k in oldKeyMap.keys():
        if k in list(parDict.keys()) and oldKeyMap[k] is None:
            del parDict[k]
            if verbose:
                print("... WARNING: config parameter '%s' is no longer used by Nemo and will be ignored." % (k))
        if k in list(parDict.keys()) and type(oldKeyMap[k]) == str:
            if verbose:
                print("... WARNING: config parameter '%s' (old usage) has been renamed to '%s' (current usage) - you may wish to update your config file." % (k, oldKeyMap[k]))
            parDict[oldKeyMap[k]]=parDict[k]
            del parDict[k]

    if verbose:
        print("... config loaded successfully")

    return parDict

#------------------------------------------------------------------------------------------------------------
class NemoConfig(object):
    """An object that manages Nemo's configuration (paths to maps, output directories, filter settings etc.).
    
    Attributes:
        parDict (:obj:`dict`): Dictionary containing the contents of the config file.
        rootOutDir (:obj:`str`): Path to the directory where all output will be written.
        filteredMapsDir (:obj:`str`): Name of the directory where filtered maps will be written.
        diagnosticsDir (:obj:`str`): Path to the directory where miscellaneous diagnostic data (e.g., filter 
            kernel plots) will be written.
        unfilteredMapsDictList (:obj:`list`): List of dictionaries corresponding to maps needed.
        tileNames (:obj:`list`): List of map tiles (extension names) to operate on.
        MPIEnabled (:obj:`bool`): If True, use MPI to divide `tileNames` list among processes.
        comm (:obj:`MPI.COMM_WORLD`): Used by MPI.
        rank (:obj:`int`): Used by MPI.
        size (:obj:`int`): Used by MPI.
    
    """
    
    def __init__(self, config, makeOutputDirs = True, setUpMaps = True, writeTileInfo = False,
                 selFnDir = None, calcSelFn = False, sourceInjectionTest = False, MPIEnabled = False,
                 divideTilesByProcesses = True, verbose = True, strictMPIExceptions = True):
        """Creates an object that manages Nemo's configuration (paths to maps, output directories,
        filter settings etc.).
        
        Args:
            config (:obj:`str` or :obj:`dict`): Either the path to a nemo .yml configuration
                file, or a dictionary containing nemo configuration parameters.
            makeOutputDirs (:obj:`bool`, optional): If True, create output directories (where maps, 
                catalogs are stored).
            setUpMaps (:obj:`bool`, optional): If True, set-up data structures for handling maps 
                (inc. breaking into tiles if wanted).
            writeTileInfo (:obj:`bool`, optional): If True and set-up to break maps into tiles, write
                info on the tile geometry (such as a DS9 region file) into `selFnDir`.
            selFnDir (:obj:`str`, optional): Path to the selFn directory (use to override the 
                default location).
            calcSelFn (:obj:`bool`, optional): Overrides the value given in the config file if True.
            sourceInjectionTest (:obj:`bool`, optional): Overrides the value given in the config file
                if True.
            MPIEnabled (:obj:`bool`, optional): If True, use MPI to divide the map into tiles, 
                distributed among processes. This requires `tileDefinitions` and `tileNoiseRegions` 
                to be given in the .yml config file.
            divideTilesByProcesses (:obj:`bool`, optional): If True, divides up the list of tiles
                optimally among the available MPI processes.
            strictMPIExceptions (:obj:`bool`): If True, MPI will abort if an Exception is encountered
                (the downside is that you will not get the full traceback, but at least you will not waste
                CPU cycles). If False, MPI probably will not abort if an Exception is encountered, but you 
                will get the full traceback (the downside is your MPI job may never complete). These options 
                are a compromise due to how mpi4py handles MPI errors (the default handling for mpi4py 
                corresponds to strictMPIExceptions = False).
            verbose (:obj:`bool`): If True, print some info to the terminal while we set-up the config file.
    
        """
        self.MPIEnabled=MPIEnabled
        if self.MPIEnabled == True:
            from mpi4py import MPI
            # This is needed to get MPI to abort if one process crashes (due to mpi4py error handling)
            # If this part is disabled, we get nice exceptions, but the program will never finish if a process dies
            # Here we get the error message at least but not the traceback before MPI Aborts
            if strictMPIExceptions == True:
                sys_excepthook=sys.excepthook
                def mpi_excepthook(v, t, tb):
                    sys_excepthook(v, t, tb)
                    print("Exception: %s" % (t.args[0]))
                    MPI.COMM_WORLD.Abort(1)
                sys.excepthook=mpi_excepthook
            self.comm=MPI.COMM_WORLD
            self.size=self.comm.Get_size()
            self.rank=self.comm.Get_rank()
            if self.size == 1:
                raise Exception("If you want to use MPI, run with e.g., mpiexec -np 4 nemo configFile.yml -M")
        else:
            self.rank=0
            self.comm=None
            self.size=1

        if self.rank != 0:
            self.verbose=False
        else:
            self.verbose=verbose

        # Timekeeping for benchmarking
        if self.rank == 0:
            self._timeStarted=time.time()

        if type(config) == str:
            self.parDict=parseConfigFile(config, verbose = self.verbose)
            self.configFileName=config
        elif type(config) == dict:
            self.parDict=config
            self.configFileName=''
        else:
            raise Exception("'config' must be either a path to a .yml file, or a dictionary of parameters.")
        
        # Handle a couple of optional command-line args. These only override if set to True, otherwise ignored
        if calcSelFn == True:
            self.parDict['calcSelFn']=True
        if sourceInjectionTest == True:
            self.parDict['sourceInjectionTest']=True

        # We want the original map WCS and shape (for using stitchMaps later)
        try:
            with pyfits.open(self.parDict['unfilteredMaps'][0]['mapFileName']) as img:
                # NOTE: Zapping keywords here that appear in old ACT maps but which confuse astropy.wcs
                # Also handling compressed maps
                for ext in img:
                    if img[ext].data is not None:
                        break
                self.origWCS=astWCS.WCS(img[ext].header, mode = 'pyfits', zapKeywords = ['PC1_1', 'PC1_2', 'PC2_1', 'PC2_2'])
                self.origShape=(img[ext].header['NAXIS2'], img[ext].header['NAXIS1'])
        except:
            # We don't always need or want this... should we warn by default if not found?
            self.origWCS=None
            self.origShape=None
                
        # We keep a copy of the original parameters dictionary in case they are overridden later and we want to
        # restore them (e.g., if running source-free sims).
        self._origParDict=copy.deepcopy(self.parDict)
                                
        # Output dirs
        self.rootOutDir=None
        if 'outputDir' in list(self.parDict.keys()):
            self.rootOutDir=os.path.abspath(self.parDict['outputDir'])
            # if os.path.exists(self.rootOutDir) == False: This isn't helpful really
            #     self.rootOutDir=None
        if self.rootOutDir is None:
            if self.configFileName.find(".yml") == -1 and makeOutputDirs == True:
                raise Exception("File must have .yml extension")
            self.rootOutDir=os.getcwd()+os.path.sep+os.path.split(self.configFileName.replace(".yml", ""))[-1]
        self.filteredMapsDir=self.rootOutDir+os.path.sep+"filteredMaps"
        self.diagnosticsDir=self.rootOutDir+os.path.sep+"diagnostics"
        self.selFnDir=self.rootOutDir+os.path.sep+"selFn"
        dirList=[self.rootOutDir, self.diagnosticsDir, self.filteredMapsDir, self.selFnDir]
        madeOutputDirs=None
        if self.rank == 0 and makeOutputDirs == True:
            for d in dirList:
                os.makedirs(d, exist_ok = True)
            madeOutputDirs=True

        # Optional override of selFn directory location
        if selFnDir is not None:
            self.selFnDir=selFnDir

        if setUpMaps == True:
            if self.verbose == True:
                print(">>> Setting up maps")
            self._setUpMaps(writeTileInfo = writeTileInfo)
        else:
            pickleFileName=self.selFnDir+os.path.sep+"tileCoordsDict.pkl"
            if os.path.exists(pickleFileName) == False:
                raise Exception("You can only use setUpMaps = False if a previous run has created the file %s" % (pickleFileName))
            with open(pickleFileName, "rb") as pickleFile:
                unpickler=pickle.Unpickler(pickleFile)
                self.tileCoordsDict=unpickler.load()
                self.tileNames=self.tileCoordsDict.keys()

        # For when we want to test on only a subset of tiles
        if 'tileNameList' in list(self.parDict.keys()):
            newList=[]
            for name in self.tileNames:
                if name in self.parDict['tileNameList']:
                    newList.append(name)
            if newList == []:
                raise Exception("tileNameList given in nemo config file but no extensions in images match")
            self.tileNames=newList
        else:
            # If we don't have / didn't set-up maps, we would still want the list of tile names
            if os.path.exists(self.selFnDir+os.path.sep+"tileCoordsDict.pkl") == True:
                with open(self.selFnDir+os.path.sep+"tileCoordsDict.pkl", "rb") as pickleFile:
                    unpickler=pickle.Unpickler(pickleFile)
                    tileCoordsDict=unpickler.load()
                assert(tileCoordsDict != {})
                self.tileCoordsDict=tileCoordsDict
                self.tileNames=list(tileCoordsDict.keys())

        # For convenience, keep the full list of tile names
        # (for when we don't need to be running in parallel - see, e.g., signals.getFRelWeights)
        self.allTileNames=self.tileNames.copy()

        # MPI: Divide up number of pixels to process as evenly as we can
        # For stuff with weight < 0.5: assign two tiles per process, otherwise one
        if self.MPIEnabled == True and divideTilesByProcesses == True:
            totalPix=0
            tilePix=np.zeros(len(self.tileNames))
            index=0
            for key in self.tileNames:
                tilePix[index]=self.tileCoordsDict[key]['numPix']
                totalPix=totalPix+tilePix[index]
                index=index+1
            targetPixPerProcess=totalPix/self.size
            pixTab=atpy.Table()
            pixTab['tileName']=self.tileNames
            pixTab['numPix']=tilePix
            pixTab['weight']=pixTab['numPix']/max(pixTab['numPix'])
            pixTab.sort('numPix')
            # Optimal number - if scaling isn't an issue [+1 because rank 0 doesn't process tiles]
            balancedNumProcesses=1+int(round((pixTab['weight'] <= 0.5).sum()/2))+(pixTab['weight'] > 0.5).sum()
            rankExtNames={}
            rankCounter=1
            weightCounter=0
            weightLimit=0.51
            for e in self.tileNames:
                if rankCounter not in rankExtNames:
                    rankExtNames[rankCounter]=[]
                rankExtNames[rankCounter].append(e)
                weightCounter=weightCounter+pixTab['weight'][pixTab['tileName'] == e]
                if weightCounter > weightLimit:
                    rankCounter=rankCounter+1
                    weightCounter=0
                if rankCounter > self.size-1:
                    rankCounter=1
            if self.rank in rankExtNames.keys():
                self.tileNames=rankExtNames[self.rank]
            else:
                self.tileNames=[]
            if self.rank == 0 and verbose == True:
                print(">>> Total tiles = %d ; total processes = %d ; balanced number of processes = %d" % (len(self.allTileNames), self.size, balancedNumProcesses))

        # # MPI: just divide up tiles pointed at by tileNames among processes
        # if self.MPIEnabled == True and divideTilesByProcesses == True:
        #     # New - bit clunky but distributes more evenly
        #     rankExtNames={}
        #     rankCounter=1
        #     for e in self.tileNames:
        #         if rankCounter not in rankExtNames:
        #             rankExtNames[rankCounter]=[]
        #         rankExtNames[rankCounter].append(e)
        #         rankCounter=rankCounter+1
        #         if rankCounter > self.size-1:
        #             rankCounter=1
        #     if self.rank in rankExtNames.keys():
        #         self.tileNames=rankExtNames[self.rank]
        #     else:
        #         self.tileNames=[]
        # if self.rank == 0:
        #     print(">>> Total tiles = %d ; total processes = %d" % (len(self.allTileNames), self.size))

        # We're now writing items per tile into their own dir (friendlier for Lustre)
        # NOTE: No longer writing individual tile filtered maps - only stitched versions
        if makeOutputDirs == True:
            for tileName in self.tileNames:
                for d in [self.diagnosticsDir, self.selFnDir]:
                    os.makedirs(d+os.path.sep+tileName, exist_ok = True)

        # Identify filter sets, for enabling new multi-pass filtering and object finding
        self._identifyFilterSets()

        # Cache filters in memory to save some I/O
        self.cachedFilters={}
        
        # For debugging...
        if self.MPIEnabled == True and verbose == True:
            print((">>> rank = %d [PID = %d]: tileNames = %s" % (self.rank, os.getpid(), str(self.tileNames))))
  
  
    def _identifyFilterSets(self):
        """Inspect the config dictionary to identify filter sets, which are used in multi-pass map filtering
        and object detection.

        """
        self.filterSets=[]
        if 'filterSetOptions' in self.parDict.keys():
            self.filterSetOptions=self.parDict['filterSetOptions']
            for filtDict in self.parDict['mapFilters']:
                if 'filterSets' in filtDict.keys():
                    for f in filtDict['filterSets']:
                        if f not in self.filterSets:
                            self.filterSets.append(f)
            self.filterSets.sort()
            self.filterSetLabels={}
            for setNum in self.filterSetOptions.keys():
                if 'label' in self.filterSetOptions[setNum].keys():
                    self.filterSetLabels[setNum]=self.filterSetOptions[setNum]['label']
                else:
                    self.filterSetLabels[setNum]=None


    def addAutoTileDefinitions(self, DS9RegionFileName = None, cacheFileName = None):
        """Runs the autotiler to add automatic tile definitions into the parameters dictionary in-place.

        Args:
            DS9RegionFileName (str, optional): Path to DS9 regions file to be written.
            cacheFileName (str, optional): Path to output a cached .yml file which will can be read instead on
                repeated runs (for speed).

        """

        # This is a bad idea if we're doing re-runs with -T to check tiling
        # if cacheFileName is not None and os.path.exists(cacheFileName):
        #     with open(cacheFileName, "r") as stream:
        #         self.parDict['tileDefinitions']=yaml.safe_load(stream)
        #     return None

        if 'tileDefinitions' in self.parDict.keys() and type(self.parDict['tileDefinitions']) == dict:
            # If we're not given a survey mask, we'll make one up from the map image itself
            if 'mask' in self.parDict['tileDefinitions'].keys() and self.parDict['tileDefinitions']['mask'] is not None:
                surveyMaskPath=self.parDict['tileDefinitions']['mask']
                maskProvided=True
            else:
                surveyMaskPath=self.parDict['unfilteredMaps'][0]['mapFileName']
                maskProvided=False
            if maskProvided == True:
                surveyMask, wcs=maps.chunkLoadMask(surveyMaskPath) # memory efficient
            else:
                with pyfits.open(surveyMaskPath, do_not_scale_image_data = True) as img:
                    for ext in img:
                        if ext.data is not None:
                            break
                    surveyMask=np.array(ext.data, dtype = np.uint8)
                    wcs=astWCS.WCS(ext.header, mode = 'pyfits')
                    # One day we will write a routine to deal with the multi-plane thing sensibly...
                    # But today is not that day
                    if surveyMask.ndim == 3:
                        surveyMask=surveyMask[0, :]
                    assert(surveyMask.ndim == 2)
                    surveyMask[surveyMask != 0]=1
                del img
            self._tileDefinitionsMaskPath=surveyMaskPath
            self.parDict['tileDefinitions']=maps.autotiler(surveyMask, wcs,
                                                           self.parDict['tileDefinitions']['targetTileWidthDeg'],
                                                           self.parDict['tileDefinitions']['targetTileHeightDeg'])
            if self.verbose:
                print("... breaking map into %d tiles" % (len(self.parDict['tileDefinitions'])))

            if DS9RegionFileName is not None:
                maps.saveTilesDS9RegionsFile(self.parDict, DS9RegionFileName)

            if cacheFileName is not None:
                stream=yaml.dump(self.parDict['tileDefinitions'])
                with open(cacheFileName, "w") as outFile:
                    outFile.write(stream)


    def getTileCoordsDict(self):
        """Construct a dictionary that describes how a large map is broken up into smaller tiles
        (see :ref:`Tiling` for information on the relevant configuration file parameters).

        Returns:
            A dictionary indexed by tile name, where each entry is a dictionary containing information
            on pixel coordinates of each tile within the larger map, and the WCS of each tile.

        """
        # Spin through a map, figuring out the actual coords to clip based on the tile definitions
        clipCoordsDict={}

        # We can take any map, because we earlier verified they have consistent WCS and size
        wcs=None
        try:
            wcsPath=self._tileDefinitionsMaskPath # For set-up of tiling with just a mask, nemo-sim-kit related
        except:
            wcsPath=self.parDict['unfilteredMaps'][0]['mapFileName'] # The old behavior
        with pyfits.open(wcsPath) as img:
            for ext in img:
                if ext.data is not None:
                    break
            wcs=astWCS.WCS(ext.header, mode = 'pyfits')
        assert(wcs is not None)

        # Untiled
        if self.parDict['useTiling'] == False:
            clipCoordsDict[ext.name]={'clippedSection': [0, wcs.header['NAXIS1'], 0, wcs.header['NAXIS2']],
                                      'header': wcs.header,
                                      'areaMaskInClipSection': [0, wcs.header['NAXIS1'], 0, wcs.header['NAXIS2']],
                                      'reprojectToTan': self.parDict['reprojectToTan']}

        # Tiled - this takes about 4 sec
        if self.parDict['useTiling'] == True:
            if self.verbose: print(">>> Finding tile coords")
            # Extract tile definitions (may have been inserted by autotiler before calling here)
            tileNames=[]
            coordsList=[]
            for tileDict in self.parDict['tileDefinitions']:
                ra0, ra1, dec0, dec1=tileDict['RADecSection']
                x0, y0=wcs.wcs2pix(ra0, dec0)
                x1, y1=wcs.wcs2pix(ra1, dec1)
                xMin=min([x0, x1])
                xMax=max([x0, x1])
                yMin=min([y0, y1])
                yMax=max([y0, y1])
                coordsList.append([xMin, xMax, yMin, yMax])
                tileNames.append(tileDict['tileName'])
            # Define clip regions in terms of pixels, adding overlap region
            tileOverlapDeg=self.parDict['tileOverlapDeg']
            mapData=np.ones([wcs.header['NAXIS2'], wcs.header['NAXIS1']], dtype = np.uint8)
            for c, name, tileDict in zip(coordsList, tileNames, self.parDict['tileDefinitions']):
                y0=c[2]
                y1=c[3]
                x0=c[0]
                x1=c[1]
                ra0, dec0=wcs.pix2wcs(x0, y0)
                ra1, dec1=wcs.pix2wcs(x1, y1)
                # Be careful with signs here... and we're assuming approx pixel size is ok
                if x0-tileOverlapDeg/wcs.getPixelSizeDeg() > 0:
                    ra0=ra0+tileOverlapDeg
                if x1+tileOverlapDeg/wcs.getPixelSizeDeg() < mapData.shape[1]:
                    ra1=ra1-tileOverlapDeg
                if y0-tileOverlapDeg/wcs.getPixelSizeDeg() > 0:
                    dec0=dec0-tileOverlapDeg
                if y1+tileOverlapDeg/wcs.getPixelSizeDeg() < mapData.shape[0]:
                    dec1=dec1+tileOverlapDeg
                if ra1 > ra0:
                    ra1=-(360-ra1)
                clip=astImages.clipUsingRADecCoords(mapData, wcs, ra1, ra0, dec0, dec1)

                # Storing clip coords etc. so can stitch together later
                # areaMaskSection here is used to define the region that would be kept (takes out overlap)
                ra0, dec0=wcs.pix2wcs(x0, y0)
                ra1, dec1=wcs.pix2wcs(x1, y1)
                clip_x0, clip_y0=clip['wcs'].wcs2pix(ra0, dec0)
                clip_x1, clip_y1=clip['wcs'].wcs2pix(ra1, dec1)
                clip_x0=int(round(clip_x0))
                clip_x1=int(round(clip_x1))
                clip_y0=int(round(clip_y0))
                clip_y1=int(round(clip_y1))
                if name not in clipCoordsDict:
                    clipCoordsDict[name]={'clippedSection': clip['clippedSection'], 'header': clip['wcs'].header,
                                          'areaMaskInClipSection': [clip_x0, clip_x1, clip_y0, clip_y1],
                                          'reprojectToTan': self.parDict['reprojectToTan'],
                                          'numPix': int(abs(clip_x1-clip_x0)*abs(clip_y1-clip_y0))}
                    if self.verbose:
                        print("... adding %s [%d, %d, %d, %d ; %d, %d]" % (name, ra1, ra0, dec0, dec1, ra0-ra1, dec1-dec0))

        return clipCoordsDict


    def _setUpMaps(self, writeTileInfo = False):

        if self.rank == 0:

            # Check any mask files are valid (e.g., -ve values can cause things like -ve area if not caught)
            maskKeys=['surveyMask', 'pointSourceMask']
            for key in maskKeys:
                if key in self.parDict.keys() and self.parDict[key] is not None:
                    maps.checkMask(self.parDict[key])

            # Check all maps have same WCS before we start
            self._checkWCSConsistency()

            if writeTileInfo == True:
                DS9RegionFileName=self.selFnDir+os.path.sep+"tiles.reg"
                cacheFileName=self.selFnDir+os.path.sep+"tileDefinitions.yml"
            else:
                DS9RegionFileName=None
                cacheFileName=None

            self.addAutoTileDefinitions(DS9RegionFileName = DS9RegionFileName, cacheFileName = cacheFileName)
            self.tileCoordsDict=self.getTileCoordsDict()
            assert(self.tileCoordsDict != {})

            bcastParDict=self.parDict
            bcastTileCoordsDict=self.tileCoordsDict
            if writeTileInfo == True:
                with open(self.selFnDir+os.path.sep+"tileCoordsDict.pkl", "wb") as pickleFile:
                    pickler=pickle.Pickler(pickleFile)
                    pickler.dump(self.tileCoordsDict)
        else:
            bcastParDict=None
            bcastTileCoordsDict=None
        if self.MPIEnabled == True:
            bcastParDict=self.comm.bcast(bcastParDict, root = 0)
            bcastTileCoordsDict=self.comm.bcast(bcastTileCoordsDict, root = 0)
            self.comm.barrier()
        self.tileNames=list(bcastTileCoordsDict.keys())
        self.parDict=bcastParDict
        self.tileCoordsDict=bcastTileCoordsDict

        self.unfilteredMapsDictList=maps.MapDictList(self.parDict['unfilteredMaps'],
                                                     tileCoordsDict = self.tileCoordsDict)

        # We keep a copy of the original maps set-up in case we want to override later
        self._origUnfilteredMapsDictList=copy.deepcopy(self.unfilteredMapsDictList)


    def _checkWCSConsistency(self):
        # Check consistency of WCS across maps
        mapKeys=['mapFileName', 'weightsFileName', 'pointSourceMask', 'surveyMask', 'flagMask']
        refWCS=None
        for mapDict in self.parDict['unfilteredMaps']:
            for key in mapKeys:
                if key in mapDict.keys() and mapDict[key] is not None:
                    with pyfits.open(mapDict[key]) as img:
                        wcs=None
                        for ext in img:
                            if ext is not None:
                                wcs=astWCS.WCS(img[ext].header, mode = 'pyfits')
                        if wcs is None:
                            raise Exception("Map %s doesn't have a WCS." % (mapDict[key]))
                        if refWCS is None:
                            refWCS=wcs
                        else:
                            try:
                                #assert(refWCS.getCentreWCSCoords() == wcs.getCentreWCSCoords())
                                #assert(refWCS.getImageMinMaxWCSCoords() == wcs.getImageMinMaxWCSCoords())
                                assert(refWCS.header['CTYPE1'] == wcs.header['CTYPE1'])
                                assert(refWCS.header['CTYPE2'] == wcs.header['CTYPE2'])
                                assert(refWCS.header['NAXIS1'] == wcs.header['NAXIS1'])
                                assert(refWCS.header['NAXIS2'] == wcs.header['NAXIS2'])
                                assert(refWCS.getXPixelSizeDeg() == wcs.getXPixelSizeDeg())
                                assert(refWCS.getYPixelSizeDeg() == wcs.getYPixelSizeDeg())
                            except:
                                raise Exception("WCS of %s is not consistent with other maps (all maps and masks must have the same WCS)." % (mapDict[key]))


    def restoreConfig(self):
        """Restores the parameters dictionary (self.parDict) and unfiltered maps list to the original
        state specified in the config .yml file.
        
        """      
        self.parDict=copy.deepcopy(self._origParDict)
        self.unfilteredMapsDictList=copy.deepcopy(self._origUnfilteredMapsDictList)


    def setFilterSet(self, setNum):
        """For use with multi-pass mode, in conjunction with filterSets and filterSetOptions config
        file parameters.

        """
        self.restoreConfig()

        options=None
        if setNum in self.filterSetOptions.keys():
            options=self.filterSetOptions[setNum]
            if 'saveCatalog' not in options.keys():
                options['saveCatalog']=False
            #if 'maskSubtractedRegions' not in options.keys():
                #options['maskSubtractedRegions']=False
            if 'maskHoleDilationFactor' not in options.keys():
                options['maskHoleDilationFactor']=None
            if 'addSiphonedFromSets' not in options.keys():
                options['addSiphonedFromSets']=None
            if 'ignoreSurveyMask' not in options.keys():
                options['ignoreSurveyMask']=False

        # We could add some checks here for options that don't make sense
        # e.g., if addSiphonedFromSets is present, better have subtractModelFromSets present

        permittedOverrides=['thresholdSigma', 'objIdent', 'findCenterOfMass', 'measureShapes']
        if options is not None:
            for override in permittedOverrides:
                if override in options.keys():
                    self.parDict[override]=options[override]

        # Set the filters and add catalogs used for model subtraction from filter noise term
        saveKeys=['saveFilteredMaps', 'saveFilter', 'saveRMSMap', 'savePlots', 'saveDS9Regions']
        filtersToActivate=[]
        for filtDict in self.parDict['mapFilters']:
            if setNum in filtDict['filterSets']:
                if options is not None:
                    if 'mapToUse' in options.keys():
                        filtDict['params']['mapToUse']=options['mapToUse']
                    if 'noiseModelCatalogFromSets' in options.keys():
                        filtDict['params']['noiseModelCatalog']=[]
                        for noiseSubIndex in options['noiseModelCatalogFromSets']:
                            filtDict['params']['noiseModelCatalog'].append(self.filterSetOptions[noiseSubIndex]['catalog'])
                # NOTE: We prevent any filter set apart from the last one from writing maps, filters to disk
                # Similarly, no point doing forced photometry on any run except the last
                if setNum != self.filterSets[-1]:
                    for saveKey in saveKeys:
                        if saveKey in filtDict['params'].keys():
                            filtDict['params'][saveKey]=False
                    self.parDict['forcedPhotometryCatalog']=None
                # However, we do allow intermediate filter sets to write maps, if explicitly asked for
                # (e.g., useful for debugging a point source run before a cluster run)
                if type(options) == dict and 'saveFilteredMaps' in options.keys():
                    filtDict['params']['saveFilteredMaps']=options['saveFilteredMaps']
                filtersToActivate.append(filtDict)
        self.parDict['mapFilters']=filtersToActivate

        # Add catalogs for model subtraction from maps
        if options is not None and 'subtractModelFromSets' in options.keys():
            for mapDict in self.unfilteredMapsDictList:
                for subtractSetIndex in options['subtractModelFromSets']:
                    if 'mapToUse' in self.filterSetOptions[subtractSetIndex].keys():
                        if 'label' in mapDict.keys() and mapDict['label'] != self.filterSetOptions[subtractSetIndex]['mapToUse']:
                            continue
                    mapDict['subtractModelFromCatalog']=self.filterSetOptions[subtractSetIndex]['catalog']

        # Add catalogs for masking and filling
        if options is not None and 'maskAndFillFromSets' in options.keys():
            for mapDict in self.unfilteredMapsDictList:
                for subtractSetIndex in options['maskAndFillFromSets']:
                    if 'mapToUse' in self.filterSetOptions[subtractSetIndex].keys():
                        if 'label' in mapDict.keys() and mapDict['label'] != self.filterSetOptions[subtractSetIndex]['mapToUse']:
                            continue
                    mapDict['maskAndFillFromCatalog']=self.filterSetOptions[subtractSetIndex]['catalog']

        # Other map-level preprocessing keys
        if options is not None:
            for mapDict in self.unfilteredMapsDictList:
                #mapDict['maskSubtractedRegions']=options['maskSubtractedRegions']
                mapDict['maskHoleDilationFactor']=options['maskHoleDilationFactor']
                if options['ignoreSurveyMask'] == True:
                    mapDict['surveyMask']=None



"""

Code to churn out a Gaussian beam in a format that nemo understands

"""

import os
import sys
import argparse
import numpy as np
import IPython

parser = argparse.ArgumentParser()
parser.add_argument("FWHMArcmin")
parser.add_argument("outFileName")
args = parser.parse_args()

FWHMArcmin=float(args.FWHMArcmin)
sigmaDeg=(FWHMArcmin/(2*np.sqrt(2*np.log(2))))/60. 

# Radial distance in degrees
xRange=np.linspace(0.0, 0.5, 1800)

# Gaussian
resp=np.exp(-np.power(xRange, 2)/(2*np.power(sigmaDeg, 2)))

# Solid angle in nsr
solidAngle=np.trapz(resp, np.pi*(np.radians(xRange))**2)*1e9

# Output format: plain text; 1st column degrees; 2nd column response (normalised to 1.0 at 0.00)
with open(args.outFileName, "w") as outFile:
    outFile.write("# solid angle = %.2f nsr\n" % (solidAngle))
    for x, r in zip(xRange, resp):
        outFile.write("%.6f %.6e\n" % (x, r))



"""

Generate .yml configs for various maps and beams.

It's too complicated to understand MFMF maps (because so many pa and set combos)

So, let's start with single frequency only

This runs quick enough that we can stick everything in one batch job

"""

import os
import sys
import numpy as np
import IPython

# Map calibration factors (Steve, via Sigurd ACTPol Slack) --------------------------------------------------
# NOTE: These are just map * factor (pre-S17 needed sqrt, which has already been applied here)
mapCalibFactors={"s13_pa1_f150_deep1":  np.sqrt(0.974),
                 "s13_pa1_f150_deep5":  np.sqrt(1.113),
                 "s13_pa1_f150_deep6":  np.sqrt(0.942),
                 "s14_pa1_f150_deep56": np.sqrt(0.997),
                 "s14_pa2_f150_deep56": np.sqrt(0.978),
                 "s15_pa1_f150_deep56": np.sqrt(0.987),
                 "s15_pa1_f150_boss":   np.sqrt(1.070),
                 "s15_pa1_f150_deep8":  np.sqrt(1.023),
                 "s15_pa2_f150_deep56": np.sqrt(0.958),
                 "s15_pa2_f150_boss":   np.sqrt(1.063),
                 "s15_pa2_f150_deep8":  np.sqrt(1.099),
                 "s15_pa3_f090_deep56": np.sqrt(1.059),
                 "s15_pa3_f090_boss":   np.sqrt(1.125),
                 "s15_pa3_f090_deep8":  np.sqrt(1.110),
                 "s15_pa3_f150_deep56": np.sqrt(1.093),
                 "s15_pa3_f150_boss":   np.sqrt(1.219),
                 "s15_pa3_f150_deep8":  np.sqrt(1.221),
                 "s16_pa2_f150_cmb":    np.sqrt(0.951),
                 "s16_pa3_f090_cmb":    np.sqrt(1.136),
                 "s16_pa3_f150_cmb":    np.sqrt(1.960),
                 "s17_pa4_f150":        1.20995868,
                 "s17_pa4_f220":        0.94868330,
                 "s17_pa5_f090":        0.98843310,
                 "s17_pa5_f150":        1.03778611,
                 "s17_pa6_f090":        1.00000000,
                 "s17_pa6_f150":        1.02176318,
                 "s18_pa4_f150":        1.200,
                 "s18_pa4_f220":        1.081,
                 "s18_pa5_f090":        1.174,
                 "s18_pa5_f150":        1.049,
                 "s18_pa6_f090":        1.214,
                 "s18_pa6_f150":        1.186}

# Configs: maps/beams/labels --------------------------------------------------------------------------------
configs=[{'label':  "s14_deep56",
          'map':    "maps/s14_deep56_pa$PA_$BAND_nohwp_night_3pass_4way_set$SET_map.fits",
          'beam':   "beams/190809/b20190809_s14_pa$PA_$BAND_nohwp_night_beam_profile_jitter_deep56.txt",
          'calKey': "s14_pa$PA_$BAND_deep56",
          'surveyMask': "null"},
         {'label':  "s15_deep56",
          'map':    "maps/s15_deep56_pa$PA_$BAND_nohwp_night_3pass_4way_set$SET_map.fits",
          'beam':   "beams/190809/b20190809_s15_pa$PA_$BAND_nohwp_night_beam_profile_jitter_deep56.txt",
          'calKey': "s15_pa$PA_$BAND_deep56",
          'surveyMask': "null"},
         {'label':  "s15_boss",
          'map':    "maps/s15_boss_pa$PA_$BAND_nohwp_night_3pass_4way_set$SET_map.fits",
          'beam':   "beams/190809/b20190809_s15_pa$PA_$BAND_nohwp_night_beam_profile_jitter_boss.txt",
          'calKey': "s15_pa$PA_$BAND_boss",
          'surveyMask': "null"},
         {'label':  "s16_cmb",
          'map':    "maps/s16_cmb_pa$PA_$BAND_nohwp_night_3pass_2way_set$SET_map.fits",
          'beam':   "beams/190809/b20190809_s16_pa$PA_$BAND_nohwp_night_beam_profile_jitter_cmb.txt",
          'calKey': "s16_pa$PA_$BAND_cmb",
          'surveyMask': "surveyMask_v7_S18.fits"},
         {'label':  "s17_cmb",
          'map':    "maps/s17_cmb_pa$PA_$BAND_nohwp_night_1pass_2way_set$SET_map.fits",
          'beam':   "beams/s17_pa$PA_$BAND_nohwp_night_beam_profile_jitter.txt",
          'calKey': "s17_pa$PA_$BAND",
          'surveyMask': "surveyMask_v7_S18.fits"},
         {'label':  "s18_cmb",
          'map':    "maps/s18_cmb_pa$PA_$BAND_nohwp_night_1pass_2way_set$SET_map.fits",
          'beam':   "beams/s17_pa$PA_$BAND_nohwp_night_beam_profile_jitter.txt",
          'calKey': "s18_pa$PA_$BAND",
          'surveyMask': "surveyMask_v7_S18.fits"},]

# Make Nemo .yml configs and slurm batch script -------------------------------------------------------------         
batchLines=[]
for c in configs:
    for band, freq in zip(['f150', 'f090'], [145.3, 94.1]):
        for paNum in range(1, 7):
            for setNum in range(0, 2):
                with open("template_singlefreq.yml") as inFile:
                    lines=inFile.readlines()
                mapName=c['map'].replace("$PA", str(paNum)).replace("$SET", str(setNum)).replace("$BAND", band)
                #weightName=mapName.replace("_map.fits", "_div.fits")
                beamName=c['beam'].replace("$PA", str(paNum)).replace("$BAND", band)
                if os.path.exists(mapName) == True and os.path.exists(beamName) == True:
                    calibFactor=mapCalibFactors[c['calKey'].replace("$PA", str(paNum)).replace("$BAND", band)]
                    surveyMask=c['surveyMask']
                    outFileName=c['label']+"_%s_pa%d_set%d.yml" % (band, paNum, setNum)
                    with open(outFileName, "w") as outFile:
                        for line in lines:
                            if line.find("$MAP") != -1:
                                outFile.write(line.replace("$MAP", mapName))
                            elif line.find("$FREQ") != -1:
                                outFile.write(line.replace("$FREQ", str(freq)))
                            elif line.find("$BEAM") != -1:
                                outFile.write(line.replace("$BEAM", beamName))
                            elif line.find("$CALIBFACTOR") != -1:
                                outFile.write(line.replace("$CALIBFACTOR", str(calibFactor)))
                            elif line.find("$SURVEYMASK") != -1:
                                outFile.write(line.replace("$SURVEYMASK", surveyMask))
                            else:
                                outFile.write(line)
                    batchLines.append("time mpiexec nemo %s -M" % (outFileName))

with open("clustercal.sh", "w") as batchFile:
    batch="""#!/bin/sh
#SBATCH --nodes=18
#SBATCH --ntasks-per-node=16
#SBATCH --mem=64000
#SBATCH --time=23:59:00

source ~/.bashrc
"""
    batchFile.write(batch)
    for line in batchLines:
        batchFile.write(line+"\n")



# -*- coding: iso-8859-1 -*-
#
# nemo install script

import os
import glob
from setuptools import setup
from setuptools import Extension
import numpy
import versioneer

cmdclass=versioneer.get_cmdclass()

setup(name='nemo-sz',
      version=versioneer.get_version(),
      cmdclass=cmdclass,
      url="https://nemo-sz.readthedocs.io",
      author='Matt Hilton + Nemo Contributors',
      author_email='matt.hilton@wits.ac.za',
      packages=['nemo'],
      package_data={'nemo': ['data/*']},
      scripts=['bin/nemo', 'bin/nemoMass', 'bin/nemoMock', 'bin/nemoCatalogCheck', 'bin/nemoMask', 'bin/nemoModel', 'bin/nemoSpec'],
)



"""

Library for running nemo tests using Robot Framework

"""

import os
import sys
import subprocess
import shutil
import numpy as np
from nemo import catalogs, maps, plotSettings
from astLib import *
import astropy.io.fits as pyfits
import astropy.table as atpy
from astropy.coordinates import SkyCoord, match_coordinates_sky
import pylab as plt

plotSettings.update_rcParams()
plotTitleSize=14

#------------------------------------------------------------------------------------------------------------
class NemoTests(object):

    def __init__(self):
        """Basic set-up for running tests. For those that need datasets to be downloaded, a setup_<name> 
        routine should be added (see, e.g., setup_equD56).
        
        """

        self._status = ''

        self.cacheDir="testsCache"
        if os.path.exists(self.cacheDir) == False:
            os.makedirs(self.cacheDir)
        
        # Why don't we just run all tests from the cache dir?
        self.runDir=os.path.abspath(self.cacheDir)
            
        self.plotsDir="plots"
        if os.path.exists(self.plotsDir) == False:
            os.makedirs(self.plotsDir)
    
    
    def setup_equD56(self):
        """Set-up for tests that use E-D56 maps - downloads them if not found.
        
        """
    
        # Needed for inject_sources
        self.beamFileName=self.cacheDir+os.path.sep+"profiles_ACT/profile_AR1_2009_pixwin_130224.txt"
        
        thisDir=os.getcwd()
        
        # Download the E-D56 map if not found
        self.inMapFileName=self.cacheDir+os.path.sep+"weightedMap_4.fits"
        if os.path.exists(self.inMapFileName) == False:
            print(">>> Downloading E-D56 map ...")
            os.chdir(self.cacheDir)
            os.system("wget https://astro.ukzn.ac.za/~mjh/equD56Maps.tar.gz")
            os.system("tar -zxvf equD56Maps.tar.gz")
            os.remove("equD56Maps.tar.gz")
            os.chdir(thisDir)
        
        # This one is actually in the distribution as it stands but anyway...
        if os.path.exists(self.cacheDir+os.path.sep+"E-D56Clusters.fits") == False:
            os.chdir(self.cacheDir)
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/actpol_2016_lensing/E-D56Clusters.fits")
            os.chdir(thisDir)
            
        # Copy files that are in the git/source distribution (saves re-organising)
        if os.path.exists(self.cacheDir+os.path.sep+"surveyMask.fits.gz") == False:
            os.system("cp %s %s/" % ("../examples/ACT-DR3-clusters/surveyMask.fits.gz", self.cacheDir))
        if os.path.exists(self.cacheDir+os.path.sep+"pointSourceMask.fits.gz") == False:
            os.system("cp %s %s/" % ("../examples/ACT-DR3-clusters/pointSourceMask.fits.gz", self.cacheDir))
        if os.path.exists(self.cacheDir+os.path.sep+"ACTPol_redshifts.fits") == False:
            os.system("cp %s %s/" % ("../examples/ACT-DR3-clusters/ACTPol_redshifts.fits", self.cacheDir))
        if os.path.exists(self.cacheDir+os.path.sep+"profiles_ACT") == False:
            os.system("cp -R %s %s/" % ("../examples/ACT-DR3-clusters/profiles_ACT", self.cacheDir))
            
        
    def setup_south2008(self):
        """Set-up for tests that use southern 2008 ACT data - downloads needed files from LAMBDA if not 
        found.
        
        """
        
        thisDir=os.getcwd()
        self.inMapFileName=self.cacheDir+os.path.sep+"ACT_148_south_season_2_1way_v3_summed.fits"
        if os.path.exists(self.inMapFileName) == False:
            print(">>> Downloading South 2008 data ...")
            os.chdir(self.cacheDir)
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/data2013/Maps/AR1/South/ACT_148_south_season_2_1way_v3_summed.fits")
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/data2013/Weights/AR1/South/ACT_148_south_season_2_1way_hits_v3.fits")
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/data2013/Beams/profiles/profile_AR1_2008_pixwin_130224.txt")
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/Cluster_src/Ptsrc_cat/act_source_catalog_AR1_2008.txt")
            os.chdir(thisDir)
            
        # Need to convert published catalog such that comparison routines work
        # NOTE: loading the table this way breaks the name column but we soldier on...
        tabFileName=self.cacheDir+os.path.sep+"act_source_catalog_AR1_2008.fits"
        if os.path.exists(tabFileName) == False:
            tab=atpy.Table().read(self.cacheDir+os.path.sep+"act_source_catalog_AR1_2008.txt", format = 'ascii')
            tab.rename_column("col2", "name")
            tab.rename_column("col3", "RADeg")
            tab.rename_column("col4", "decDeg")
            tab.rename_column("col6", "fluxJy")
            tab['fluxJy']=tab['fluxJy']/1000.0
            tab.write(tabFileName)


    def setup_quickstart(self):
        """Set-up for tests that use the files from the quickstart tutorial - downloads them if not found.
        
        """
    
        # Maps and beams
        thisDir=os.getcwd()        
        self.inMapFileName=self.cacheDir+os.path.sep+"maps"+os.path.sep+"f150_1_10_8_map.fits"
        if os.path.exists(self.inMapFileName) == False:
            print(">>> Downloading quickstart files ...")
            os.chdir(self.cacheDir)
            os.system("wget 'https://www.dropbox.com/scl/fi/5sjapfshk8g3kxgy4zc56/nemo-quickstart-maps.tar.gz?rlkey=swpdttpgvkffbgr9pcmebslw2&dl=0' -O nemo-quickstart-maps.tar.gz")
            os.system("tar -zxvf nemo-quickstart-maps.tar.gz")
            os.remove("nemo-quickstart-maps.tar.gz")
            os.chdir(thisDir)
        
        # DR5 release catalog for comparing signals
        comparisonCatalog=self.cacheDir+os.path.sep+"DR5_cluster-catalog_v1.1.fits"
        if os.path.exists(comparisonCatalog) == False:
            os.chdir(self.cacheDir)
            os.system("wget https://lambda.gsfc.nasa.gov/data/suborbital/ACT/ACT_dr5/DR5_cluster-catalog_v1.1.fits")
            os.chdir(thisDir)
        
        # Generate mask file(s) for simming
        headerFileNames=["configs/smallTestSurveyMaskHeader.txt", "configs/MPITestSurveyMaskHeader.txt"]
        maskFileNames=["smallTestSurveyMask.fits", "MPITestSurveyMask.fits"]
        for headerFileName, maskFileName in zip(headerFileNames, maskFileNames):
            header=pyfits.Header().fromtextfile(headerFileName)
            wcs=astWCS.WCS(header, mode = 'pyfits')
            d=np.ones([wcs.header['NAXIS2'], wcs.header['NAXIS1']], dtype = int)
            maps.saveFITS(self.cacheDir+os.path.sep+maskFileName, d, wcs, compressionType = 'PLIO_1')
        
        # Map/frequency-related defaults
        self.bandsDict={'f150': {'beam': "maps/s16_pa2_f150_nohwp_night_beam_profile_jitter.txt",
                                 'freq': str(149.6)},
                        'f090': {'beam': "maps/s16_pa3_f090_nohwp_night_beam_profile_jitter.txt",
                                 'freq': str(97.8)}
                       }
        self.sizesDict={'small': {'mask': "smallTestSurveyMask.fits"},
                        'large': {'mask': "MPITestSurveyMask.fits"}
                       }
    

    def set_config(self, configFileName):
        """Set the config file to be used by all nemo scripts executed by tests.
        Path given here can be relative to the current working directory (it gets 
        turned into an absolute path if so).
        
        """
        self.configFileName=os.path.abspath(configFileName)
        self.selFnDir=configFileName.replace(".yml", "")+os.path.sep+"selFn"
        self.mocksDir=configFileName.replace(".yml", "")+os.path.sep+"mocks"
        
        
    def run_nemo(self, forcedPhotometryCatalog = None):
        args=['nemo', self.configFileName]
        if forcedPhotometryCatalog is not None:
            args=args+['-f', forcedPhotometryCatalog]
        self._run_command(args)


    def run_nemo_injection_test(self):
        self._run_command(["nemo", self.configFileName, "-I"])


    def run_parallel_nemo(self):
        self._run_command(["mpiexec", "-np", "4", "nemo", self.configFileName, "-M"])


    def run_nemo_mass(self, catalogFileName = None, forcedPhotometry = False):
        args=['nemoMass', self.configFileName]
        if catalogFileName is not None:
            args=args+['-c', catalogFileName]
        if forcedPhotometry == True:
            args=args+['-F']
        self._run_command(args)


    def make_sim(self, catalogFileName = None, noiseLevel = None, seed = None, band = None, size = None):
        args=['nemoModel', catalogFileName, self.sizesDict[size]['mask'], self.bandsDict[band]['beam'], 
              "sim_%s.fits" % (band), '-f', self.bandsDict[band]['freq'], 
              '-C', '-N', noiseLevel, '-S', seed]
        self._run_command(args)


    def make_parallel_sim(self, catalogFileName = None, noiseLevel = None, seed = None, band = None, size = None, profile = None):
        args=['mpiexec', '-np', '4', 'nemoModel', catalogFileName, self.sizesDict[size]['mask'], self.bandsDict[band]['beam'],
              "sim_%s.fits" % (band), '-f', self.bandsDict[band]['freq'],
              '-C', '-N', noiseLevel, '-S', seed, '-M']
        if profile is not None:
            args=args+['-p', profile]
        self._run_command(args)


    def make_signal_free_sim(self, noiseLevel = None, seed = None, band = None, size = None):
        args=['nemoModel', "pointsources-0", self.sizesDict[size]['mask'], self.bandsDict[band]['beam'],
              "signal_free_%s.fits" % (band), '-f', self.bandsDict[band]['freq'],
              '-C', '-N', noiseLevel, '-S', seed]
        self._run_command(args)
        

    def make_signal_only_sim(self, catalogFileName = None, band = None, size = None):
        args=['nemoModel', catalogFileName, self.sizesDict[size]['mask'], self.bandsDict[band]['beam'], 
              "signal_model_only_%s.fits" % (band), '-f', self.bandsDict[band]['freq']]
        self._run_command(args)
        
        
    def make_parallel_signal_only_sim(self, catalogFileName = None, band = None, size = None, profile = None):
        args=['mpiexec', '-np', '4', 'nemoModel', catalogFileName, self.sizesDict[size]['mask'], self.bandsDict[band]['beam'],
              "signal_model_only_%s.fits" % (band), '-f', self.bandsDict[band]['freq'], '-M']
        if profile is not None:
            args=args+['-p', profile]
        self._run_command(args)
    

    def run_nemo_selfn(self, configFileName = None):
        self._run_command(["nemoSelFn", configFileName])


    def run_nemo_mock(self, selFnDir = None, mocksDir = None):
        self._run_command(["nemoMock", selFnDir, mocksDir])
        

    def inject_sources_using_nemo(self, outMapFileName = "sourceInjectedMap.fits", 
                       catalogFileName = "inputSourcesCatalog.fits", numSources = 1000):
        """Injects a bunch of point sources into a map, using a vaguely plausible amplitude distribution
        based on the sources found in the deep56 map.
        
        """
        
        with pyfits.open(self.inMapFileName) as img:
            mapData=img[0].data
            if mapData.ndim > 2:
                mapData=mapData[0]
            wcs=astWCS.WCS(img[0].header, mode = 'pyfits')
        simCat=catalogs.generateRandomSourcesCatalog(mapData, wcs, numSources)
        simCat.write(catalogFileName, overwrite = True)
        #catalogs.catalog2DS9(simCat, catalogFileName.replace(".fits", ".reg"))

        modelMap=maps.makeModelImage(mapData.shape, wcs, simCat, self.beamFileName)
        astImages.saveFITS(outMapFileName, modelMap+mapData, wcs)
        
        
    def cross_match(self, inCatalogFileName, outCatalogFileName, radiusArcmin = 1.0):
        """Cross matches input and output source catalogs.
        
        """
        inTab=atpy.Table().read(inCatalogFileName)
        outTab=atpy.Table().read(outCatalogFileName)
        self.inTab, self.outTab, self.rDeg=catalogs.crossMatch(inTab, outTab, 
                                                               radiusArcmin = radiusArcmin)
    
    
    def getRADecKeys(self, tab):
        """Returns the column names in which RA, dec coords are stored, after trying a couple of variations.
        
        """
        RAKeysToTry=['ra', 'RADeg']
        decKeysToTry=['dec', 'decDeg']
        RAKey, decKey=None, None
        for key in RAKeysToTry:
            if key in tab.keys():
                RAKey=key
                break
        for key in decKeysToTry:
            if key in tab.keys():
                decKey=key
                break
        if RAKey == None or decKey == None:
            raise Exception("couldn't identify RA, dec columns in the supplied table")
        
        return RAKey, decKey
        
        
    def check_recovered_ratio(self, inKey, outKey, toleranceSigma = 1.0, expectedRatio = 1.0, SNRCut = 4,
                              SNRKey = 'fixed_SNR', errInKey = None, errOutKey = None,
                              plotLabel = None, plotsDir = "plots"):
        """Catalogs must have been cross matched before this can be run.
        Calculate the ratio of the columns pointed to by inKey, outKey. If tolerance (defined as
        1 - average ratio) is exceeded, the test is failed. SNRCut is applied to SNRKey in the output catalog.
        
        """

        inTab=self.inTab
        outTab=self.outTab
        mask=outTab[SNRKey] > SNRCut
        x=inTab[inKey]
        y=outTab[outKey]
        meanRatio=np.mean(y[mask])/np.mean(x[mask])
        bsRatios=[]
        for i in range(5000):
            indices=np.random.randint(0, len(x[mask]), len(x[mask]))
            bsX=x[mask][indices]
            bsY=y[mask][indices]
            bsRatios.append(np.mean(bsY)/np.mean(bsX))
        meanRatioErr=np.percentile(abs(bsRatios-meanRatio), 68.3)

        label="<input %s>/<output %s> = %.3f ± %.3f (%s > %.1f)" % (inKey, outKey, meanRatio, meanRatioErr, SNRKey, SNRCut)
        print("%s" % (label))

        if plotLabel is not None:
            plotMin=0.1*min([inTab[inKey].min(), outTab[outKey].min()])
            plotMax=10*max([inTab[inKey].max(), outTab[outKey].max()])
            plotRange=np.linspace(plotMin, plotMax, 100)
            plt.figure(figsize = (10, 8))
            if errInKey is not None and errOutKey is not None:
                plt.errorbar(inTab[inKey], outTab[outKey], yerr = outTab[errOutKey],
                     xerr = inTab[errInKey], elinewidth = 1.5, ecolor = '#AAAAFF',
                     fmt = 'D', ms = 6, label = None)
            else:
                plt.plot(inTab[inKey], outTab[outKey], 'D')
            plt.xlabel("input "+inKey)
            plt.ylabel("output "+outKey)
            plt.plot(plotRange, plotRange, 'k--')
            plt.xlim(plotMin, plotMax)
            plt.ylim(plotMin, plotMax)
            plt.loglog()
            plt.title(label, fontdict = {'size': plotTitleSize})
            plt.savefig(plotsDir+os.path.sep+plotLabel+"_XvY.png")
            plt.close()
        if abs((expectedRatio-meanRatio)/meanRatioErr) > toleranceSigma:
            self._status="FAILED"
        else:
            self._status="SUCCESS"
            
            
    def check_recovered_positions(self, toleranceArcsec = 12.0, SNRKey = 'SNR', SNRMax = 10.0, 
                                  plotLabel = None, plotsDir = "plots"):
        """Blah
        
        """
        inTab=self.inTab
        outTab=self.outTab
        rDeg=self.rDeg
        SNRs=outTab[SNRKey]
        medOffsetArcmin=np.median(rDeg)*60
        SNRMask=np.less(SNRs, SNRMax)
        medOffsetArcmin_SNR10=np.median(rDeg[SNRMask])*60#[SNRMask])*60
        print('... median recovered position offset = %.2f" (full sample)' % (medOffsetArcmin*60))
        label='median recovered position offset = %.2f" (SNR < 10)' % (medOffsetArcmin_SNR10*60)
        print("... %s" % (label))
        if plotLabel is not None:
            plt.figure(figsize = (10, 8))
            plt.plot(SNRs, rDeg*3600., 'r.')
            plt.semilogx()
            plt.xlabel("SNR")
            plt.ylabel('Position offset (")')
            plt.title(label, fontdict = {'size': plotTitleSize})
            plt.savefig(plotsDir+os.path.sep+plotLabel+"_posRec.png")
            plt.close()
        if medOffsetArcmin_SNR10*60 > toleranceArcsec:
            self._status="FAILED"
        else:
            self._status="SUCCESS"
    
    
    def subtract_maps(self, map1FileName, map2FileName, outMapFileName):
        with pyfits.open(map1FileName) as img1:
            d1=img1[0].data
            wcs=astWCS.WCS(img1[0].header, mode = 'pyfits')
        with pyfits.open(map2FileName) as img2:
            d2=img2[0].data
        maps.saveFITS(outMapFileName, d1-d2, wcs)


    def check_map_sigma(self, map1FileName, expectedSigma, tol = 1.0):
        with pyfits.open(map1FileName) as img1:
            d1=img1[0].data
        print("args:", map1FileName, expectedSigma)
        diff=abs(np.std(d1.flatten()) - float(expectedSigma))
        print("diff from expected sigma", diff)
        if diff > tol:
            self._status="FAILED"
        else:
            self._status="SUCCESS"
    
            
    def status_should_be(self, expected_status):
        if expected_status != self._status:
            raise AssertionError("Expected status to be '%s' but was '%s'."
                                 % (expected_status, self._status))


    def _run_command(self, args):
        thisDir=os.getcwd()
        os.chdir(self.runDir)
        print(args)
        process=subprocess.run(args, universal_newlines=True, stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
        if process.returncode != 0:
             print(process.stdout)
             raise AssertionError("Return code of '%s' is non-zero." % (str(args)))
        os.chdir(thisDir)



"""

This module defines the MockSurvey class, used for mass function calculations, obtaining de-biased cluster
mass estimates, selection function calculations, and generating mock catalogs.

"""

import os
import sys
import numpy as np
import astropy.table as atpy
import pylab as plt
import subprocess
from astropy.cosmology import FlatLambdaCDM
on_rtd=os.environ.get('READTHEDOCS', None)
if on_rtd is None:
    import pyccl as ccl
from . import signals
from . import catalogs
from . import maps
import pickle
from scipy import interpolate
from scipy import integrate
from scipy.interpolate import InterpolatedUnivariateSpline as _spline
from scipy import stats
from astLib import *
import time

# CLASS-SZ is quite a big import (tensorflow) if we don't need it
noCLASS=os.environ.get('NEMO_NOCLASSSZ', None)
if noCLASS is None:
    try:
        from classy_sz import Class
    except:
        pass

#------------------------------------------------------------------------------------------------------------
# Global settings for CLASS-SZ, if used
CLASS_SZ_SETTINGS={# 'mass function' : 'T08M200c',
                   'hm_consistency': 0,
                   # 'concentration parameter' : 'B13',
                   'B':1.,
                   'N_ncdm' : 1,
                   'N_ur' : 2.0328,
                   'm_ncdm' : 0.0,
                   'T_ncdm' : 0.71611,
                   'z_min': 1.e-3,
                   'z_max': 3.05,
                   'redshift_epsrel': 1e-6,
                   'redshift_epsabs': 1e-100,
                   'M_min': 1e13,
                   'M_max': 1e17,
                   'mass_epsrel':1e-6,
                   'mass_epsabs':1e-100,
                   'ndim_redshifts' :5000,
                   'ndim_masses' : 500,
                   'n_m_dndlnM' : 1000,
                   'n_z_dndlnM' : 1000,
                   'HMF_prescription_NCDM': 1,
                   'no_spline_in_tinker': 1,
                   'use_m500c_in_ym_relation' : 0,
                   'use_m200c_in_ym_relation' : 1,
                   'y_m_relation' : 1,
                   'output': 'dndlnM,m500c_to_m200c,m200c_to_m500c'}

#------------------------------------------------------------------------------------------------------------
class MockSurvey(object):
    """An object that provides routines calculating cluster counts (using `CCL <https://ccl.readthedocs.io/en/latest/>`_) 
    and generating mock catalogs for a given set of cosmological and mass scaling relation parameters.
    The Tinker et al. (2008) halo mass function is used (hardcoded at present, but in principle this can
    easily be swapped for any halo mass function supported by CCL).
    
    Attributes:
        areaDeg2 (:obj:`float`): Survey area in square degrees.
        zBinEdges (:obj:`np.ndarray`): Defines the redshift bins for the cluster counts.
        z (:obj:`np.ndarray`): Centers of the redshift bins.
        log10M (:obj:`np.ndarray`): Centers of the log10 mass bins for the cluster counts
            (in MSun, with mass defined according to `delta` and `rhoType`).
        a (:obj:`np.ndarray`): Scale factor (1/(1+z)).
        delta (:obj:``float`): Overdensity parameter, used for mass definition (e.g., 200, 500).
        rhoType (:obj:`str`): Density definition, either 'matter' or 'critical', used for mass definition.
        mdef (:obj:`pyccl.halos.massdef.MassDef`): CCL mass definition object, defined by `delta` and `rhoType`.
        transferFunction (:obj:`str`): Transfer function to use, as understood by CCL (e.g., 'eisenstein_hu', 
            'boltzmann_camb').
        H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
        Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
        Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
        sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
        ns (:obj:`float`): Scalar spectral index of matter power spectrum.
        volumeMpc3 (:obj:`float`): Co-moving volume in Mpc3 for the given survey area and cosmological
            parameters.
        numberDensity (:obj:`np.ndarray`): Number density of clusters (per cubic Mpc) on the 
            (z, log10M) grid.
        clusterCount (:obj:`np.ndarray`): Cluster counts on the (z, log10M) grid.
        numClusters (:obj:`float`): Total number of clusters in the survey area above the minimum mass
            limit.
        numClustersByRedshift (:obj:`np.ndarray`): Number of clusters in the survey area above the
            minimum mass limit, as a function of redshift.
    
    """
    def __init__(self, minMass, areaDeg2, zMin, zMax, H0, Om0, Ob0, sigma8, ns,
                 maxMass = 1e16, zStep = 0.01, numMassBins = 200, delta = 500, rhoType = 'critical',
                 transferFunction = 'boltzmann_camb', massFunction = 'Tinker08',
                 c_m_relation = 'Bhattacharya13', theoryCode = 'CCL'):
        """Create a MockSurvey object, for performing calculations of cluster counts or generating mock
        catalogs.
        
        Args:
            minMass (:obj:`float`): The minimum mass, in MSun. This should be set considerably lower than
                the actual survey completeness limit, otherwise completeness calculations will be wrong.
            areaDeg2 (:obj:`float`): Specifies the survey area in square degrees, which scales the
                resulting cluster counts accordingly.
            zMin (:obj:`float`): Minimum redshift for the (z, log10M) grid.
            zMax (:obj:`float`): Maximum redshift for the (z, log10M) grid.
            H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
            Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
            Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
            sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
            ns (:obj:`float`): Scalar spectral index of matter power spectrum.  
            maxMass (:obj:`float`, optional): The maximum mass, in MSun, used in mass function calculations.
            zStep (:obj:`float`, optional): Sets the linear spacing between redshift bins.
            numMassBins (:obj:`int`, optional): Number of mass bins to use.
            delta (:obj:``float`, optional): Overdensity parameter, used for mass definition (e.g., 200, 500).
            rhoType (:obj:`str`, optional): Density definition, either 'matter' or 'critical', used for mass definition.
            transferFunction (:obj:`str`, optional): Transfer function to use, as understood by CCL (e.g., 'eisenstein_hu',
                'boltzmann_camb').
            massFunction (:obj:`str`, optional): Name of the mass function to use, currently either 'Tinker08' or
                'Tinker10'. Mass function calculations are done by CCL.
            c_m_relation (:'obj:`str`, optional): Name of the concentration -- mass relation to assume, as understood by
                CCL (this may be used internally for conversion between mass definitions, as needed).
            theoryCode (:obj:`str`, optional): Either 'CCL' or 'CLASS-SZ'.

        """
        
        if areaDeg2 == 0:
            raise Exception("Cannot create a MockSurvey object with zero area")
        self.areaDeg2=areaDeg2
        self.areaSr=np.radians(np.sqrt(areaDeg2))**2

        zRange=np.arange(zMin, zMax+zStep, zStep)
        self.zBinEdges=zRange
        self.z=(zRange[:-1]+zRange[1:])/2.
        self.a=1./(1+self.z)
        self.numMassBins=numMassBins

        self.delta=delta
        self.rhoType=rhoType
        self.c_m_relation=c_m_relation
        self.mdef=ccl.halos.MassDef(self.delta, self.rhoType)
        self.transferFunction=transferFunction
        self.massFuncName=massFunction
        self.theoryCode=theoryCode
        if self.theoryCode not in ['CCL', 'CLASS-SZ']:
            raise Exception("theoryCode should be either 'CCL' or 'CLASS-SZ' - given %s" % (self.theoryCode))
        if self.theoryCode == 'CLASS_SZ':
            CLASS_SZ_SETTINGS['mass function']=self.massFuncName
        
        # Just for convenience when used elsewhere
        self.mdefLabel="M%d%s" % (self.delta, self.rhoType[0])
        
        self.H0=-1
        self.Om0=-1
        self.Ob0=-1
        self.sigma8=-1
        self.ns=-1
        self._get_new_cosmo(H0, Om0, Ob0, sigma8, ns)

        # NOTE: These are just MSun now (NOT MSun/h); always defined according to mdef
        self.log10M=np.linspace(np.log10(minMass), np.log10(maxMass), numMassBins)
        self.M=np.power(10, self.log10M)
        self.log10MBinEdges=np.linspace(self.log10M.min()-(self.log10M[1]-self.log10M[0])/2,
                                        self.log10M.max()+(self.log10M[1]-self.log10M[0])/2, len(self.log10M)+1)

        # Below is needed for Q calc when not using M500c definition (for now at least)
        if (self.delta == 500 and self.rhoType == 'critical') == False:
            self._M500cDef=ccl.halos.MassDef(500, "critical")
            self._transToM500c=ccl.halos.mass_translator(mass_in = self.mdef, mass_out = self._M500cDef,
                                                         concentration = self.c_m_relation)

        self.update(H0, Om0, Ob0, sigma8, ns)


    def setSurveyArea(self, areaDeg2):
        """Change the area of the survey to a user-specified value, updating the cluster
        counts accordingly.

        Args:
            areaDeg2 (:obj:`float`): Area of the survey in square degrees.

        """

        if areaDeg2 == 0:
            raise Exception("Cannot create a MockSurvey object with zero area")
        areaSr=np.radians(np.sqrt(areaDeg2))**2
        if areaDeg2 != self.areaDeg2:
            self.areaSr=areaSr
            self.areaDeg2=areaDeg2
            self._doClusterCount()


    def _get_new_cosmo(self, H0, Om0, Ob0, sigma8, ns):
        if ((self.H0 != H0) or (self.Om0 != Om0) or
            (self.Ob0 != Ob0) or (self.sigma8 != sigma8)):

            self.H0=H0
            self.Om0=Om0
            self.Ob0=Ob0
            self.sigma8=sigma8
            self.ns=ns

            # CCL cosmology - we may not want to entirely use CLASS-SZ, at least to start
            self.cosmoModel=ccl.Cosmology(Omega_c=Om0-Ob0,
                                          Omega_b=Ob0,
                                          h=0.01*H0,
                                          sigma8=sigma8,
                                          n_s=ns,
                                          transfer_function=self.transferFunction)

            if self.theoryCode == 'CCL':
                if self.massFuncName == 'Tinker10':
                    self.mfunc=ccl.halos.MassFuncTinker10(mass_def = self.mdef)
                elif self.massFuncName == 'Tinker08':
                    self.mfunc=ccl.halos.MassFuncTinker08(mass_def = self.mdef)

            elif self.theoryCode == 'CLASS-SZ':
                classDict={'Omega_b': Ob0, 'Omega_cdm':  Om0-Ob0, 'H0': H0, 'sigma8': sigma8,
                           'tau_reio':  0.0561, 'n_s': ns} # tau_reio not important but needs to be set
                self.cosmoCLASS=Class()
                self.cosmoCLASS.set(CLASS_SZ_SETTINGS) # Not sure if we want to update every time?
                self.cosmoCLASS.set(classDict)
                # self.cosmo.set(class_sz_ym_params)
                self.cosmoCLASS.compute_class_szfast()

            
    def update(self, H0, Om0, Ob0, sigma8, ns):
        """Recalculate cluster counts for the updated cosmological parameters given.
        
        Args:
            H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
            Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
            Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
            sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
            ns (:obj:`float`): Scalar spectral index of matter power spectrum.  
                
        """

        self._get_new_cosmo(H0, Om0, Ob0, sigma8, ns)
        
        self._doClusterCount()
        
        # For quick Q, fRel calc (these are in MockSurvey rather than SelFn as used by drawSample)
        self.theta500Splines=[]
        self.fRelSplines=[]
        self.Ez=ccl.h_over_h0(self.cosmoModel,self.a)
        self.Ez2=np.power(self.Ez, 2)
        self.DAz=ccl.angular_diameter_distance(self.cosmoModel,self.a)
        self.criticalDensity=ccl.physical_constants.RHO_CRITICAL*(self.Ez*self.cosmoModel['h'])**2
        for k in range(len(self.z)):
            # NOTE: Q fit uses theta500, as does fRel (hardcoded M500 - T relation in there)
            # This bit here may not be strictly necessary, since we don't need to map on to binning
            if self.delta == 500 and self.rhoType == "critical":
                interpLim_minLog10M500c=self.log10M.min()
                interpLim_maxLog10M500c=self.log10M.max()
            else:
                interpLim_minLog10M500c=np.log10(self._transToM500c(self.cosmoModel, self.M.min(), self.a[k]))
                interpLim_maxLog10M500c=np.log10(self._transToM500c(self.cosmoModel, self.M.max(), self.a[k]))
            zk=self.z[k]
            interpPoints=100
            fitM500s=np.power(10, np.linspace(interpLim_minLog10M500c, interpLim_maxLog10M500c, interpPoints))
            fitTheta500s=np.zeros(len(fitM500s))
            fitFRels=np.zeros(len(fitM500s))
            criticalDensity=self.criticalDensity[k]
            DA=self.DAz[k]
            Ez=self.Ez[k]
            R500Mpc=np.power((3*fitM500s)/(4*np.pi*500*criticalDensity), 1.0/3.0)    
            fitTheta500s=np.degrees(np.arctan(R500Mpc/DA))*60.0
            fitFRels=signals.calcFRel(zk, fitM500s, Ez)
            tckLog10MToTheta500=interpolate.splrep(np.log10(fitM500s), fitTheta500s)
            tckLog10MToFRel=interpolate.splrep(np.log10(fitM500s), fitFRels)
            self.theta500Splines.append(tckLog10MToTheta500)
            self.fRelSplines.append(tckLog10MToFRel)


    def _cumulativeNumberDensity(self, z):
        """Returns N > M (per cubic Mpc).
        
        """

        h=self.cosmoModel['h']
        dndlnM=self.mfunc(self.cosmoModel, self.M, 1/(1+z)) / np.log(10)
        dndM=dndlnM/self.M
        ngtm=integrate.cumtrapz(dndlnM[::-1], np.log(self.M), initial = 0)[::-1]
        
        MUpper=np.arange(np.log(self.M[-1]), np.log(10**18), np.log(self.M[1])-np.log(self.M[0]))
        extrapolator=_spline(np.log(self.M), np.log(dndlnM), k=1)
        MF_extr=extrapolator(MUpper)
        intUpper=integrate.simps(np.exp(MF_extr), dx=MUpper[2] - MUpper[1], even='first')
        ngtm=ngtm+intUpper
    
        return ngtm
    
    
    def _comovingVolume(self, z):
        """Returns co-moving volume in Mpc^3 (all sky) to some redshift z.
                
        """
        return 4.18879020479 * ccl.comoving_radial_distance(self.cosmoModel, 1./(1+z))**3

        
    def _doClusterCount(self):
        """Updates cluster count etc. after mass function object is updated.
        
        """

        assert(self.areaSr == np.radians(np.sqrt(self.areaDeg2))**2)
        self.fsky=self.areaSr/(4*np.pi) # in case it was updated

        if self.theoryCode == 'CCL':
            # Previous way - this agrees with CLASS-SZ method below to 0.1 % or something
            # zRange=self.zBinEdges
            # h = self.cosmoModel['h']
            # self.M=np.power(10, self.log10M) # in M_sun
            # norm_mfunc=1. / np.log(10)
            # numberDensity=[]
            # clusterCount=[]
            # totalVolumeMpc3=0.
            # for i in range(len(zRange)-1):
            #     zShellMin=zRange[i]
            #     zShellMax=zRange[i+1]
            #     zShellMid=(zShellMax+zShellMin)/2.
            #     dndlnM=self.mfunc(self.cosmoModel, self.M, 1./(1+zShellMid)) * norm_mfunc
            #     dndM = dndlnM / self.M
            #     n=dndM * np.gradient(self.M)
            #     numberDensity.append(n)
            #     shellVolumeMpc3=self._comovingVolume(zShellMax)-self._comovingVolume(zShellMin)
            #     shellVolumeMpc3=shellVolumeMpc3*(self.areaSr/(4*np.pi))
            #     totalVolumeMpc3+=shellVolumeMpc3
            #     clusterCount.append(n*shellVolumeMpc3)
            # numberDensity=np.array(numberDensity)
            # clusterCount=np.array(clusterCount)
            # self.volumeMpc3=totalVolumeMpc3
            # self.numberDensity=numberDensity
            # self.clusterCount=clusterCount
            # self.numClusters=np.sum(clusterCount)
            # self.numClustersByRedshift=np.sum(clusterCount, axis = 1)

            # New way (same as for CLASS-SZ below)
            # Missing: self.volumeMpc3, self.numberDensity
            dndmdz=np.zeros((self.z.shape[0], self.log10M.shape[0]))
            norm_mfunc=1. / np.log(10)
            dVdzdOmega=(((signals.CLight/1000)/self.cosmoModel['H0'])*np.power(1+self.z, 2)*self.cosmoModel.angular_diameter_distance(self.a)**2)/self.cosmoModel.h_over_h0(self.a)
            for i in range(self.a.shape[0]):
                dndlnM=self.mfunc(self.cosmoModel, self.M, self.a[i]) * norm_mfunc
                dndmdz[i]=4.*np.pi*self.fsky*dVdzdOmega[i]*dndlnM
            dndmdz=dndmdz.transpose()
            self.HMFRange=np.array([np.min(dndmdz),np.max(dndmdz)])
            dndz=np.trapz(dndmdz, x = np.log(self.M), axis = 0)
            self.numClusters=np.trapz(dndz, x = self.z)
            self.numClustersByRedshift=dndz*np.gradient(self.z)
            norm=np.sum(dndmdz)/self.numClusters # Should be good to 0.1%
            self.clusterCount=dndmdz.transpose()/norm

        elif self.theoryCode == 'CLASS-SZ':
            # Still need to implement (maybe): self.volumeMpc3, self.numberDensity
            lnms=np.log(np.power(10,  self.log10M)*self.cosmoCLASS.h())
            dndmdz = np.zeros((self.log10M.shape[0], self.z.shape[0]))
            for (im,mm) in enumerate(lnms):
                dndmdz[im,:]=4.*np.pi*self.fsky*np.vectorize(self.cosmoCLASS.get_volume_dVdzdOmega_at_z)(self.z)*np.vectorize(self.cosmoCLASS.get_dndlnM_at_z_and_M)(self.z,np.exp(mm))
            self.HMFRange=np.array([np.min(dndmdz),np.max(dndmdz)])
            dndz=np.trapz(dndmdz,x = lnms,axis = 0)
            self.numClusters=np.trapz(dndz, x = self.z)
            self.numClustersByRedshift=dndz*np.gradient(self.z)
            norm=np.sum(dndmdz)/self.numClusters # Should be good to 0.1%
            self.clusterCount=dndmdz.transpose()/norm


    def calcNumClustersExpected(self, MLimit = 1e13, zMin = 0.0, zMax = 2.0, compMz = None):
        """Calculate the number of clusters expected above a given mass limit, for the
        mass definition set when the MockSurvey object was constructed.
        
        Args:
            MLimit (:obj:`float`, optional): Mass limit above which to count clusters, in MSun.
            zMin (:obj:`float`, optional): Count clusters above this minimum redshift.
            zMax (:obj:`float`, optional): Count clusters below this maximum redshift.
            compMz (:obj:`np.ndarray`, optional): If given, a 2d array with the same dimensions
                and binning as the (z, log10M) grid, as calculated by the
                :class:`nemo.completeness.SelFn` class, that describes the completeness as
                a function of mass and redshift.
        
        Returns:
            The number of clusters in the survey area, according to the chose sample selection
            cuts.
        
        """
        
        if type(compMz) == np.ndarray:
            numClusters=compMz*self.clusterCount
        else:
            numClusters=self.clusterCount
        
        zMask=np.logical_and(np.greater(self.z, zMin), np.less(self.z, zMax))
        mMask=np.greater(self.M, MLimit)
        
        return numClusters[:, mMask][zMask].sum()
        

    def getPLog10M(self, z):
        """Returns the log10(mass) probability distribution at the given z, for the logarithmic mass
        binning and mass definition set when the MockSurvey object was constructed.
        
        Args:
            z (:obj:`float`): Redshift at which to calculate P(log10(M)).
        
        Returns:
            Array corresponding to the log10(mass) probability distribution.
        
        """
        numberDensity=self._cumulativeNumberDensity(z)
        PLog10M=numberDensity/np.trapz(numberDensity, self.M)

        return PLog10M


    def drawSample(self, y0Noise, scalingRelationDict, QFit = None, wcs = None, photFilterLabel = None,\
                   tileName = None, SNRLimit = None, makeNames = False, z = None, numDraws = None,\
                   areaDeg2 = None, applySNRCut = False, applyPoissonScatter = True,\
                   applyIntrinsicScatter = True, applyNoiseScatter = True,\
                   applyRelativisticCorrection = True, verbose = False, biasModel = None):
        """Draw a cluster sample from the mass function, generating mock y0~ values (called `fixed_y_c` in
        Nemo catalogs) by applying the given scaling relation parameters, and then (optionally) applying
        a survey selection function.
        
        Args:
            y0Noise (:obj:`float` or :obj:`np.ndarray`): Either a single number (if using e.g., a survey
                average), an RMS table (with columns 'areaDeg2' and 'y0RMS'), or a noise map (2d array).
                A noise map must be provided here if you want the output catalog to contain RA, dec
                coordinates (in addition, a WCS object must also be provided - see below).
            scalingRelationDict (:obj:`dict`): A dictionary containing keys 'tenToA0', 'B0', 'Mpivot',
                'sigma_int' that describes the scaling relation between y0~ and mass (this is the
                format of `massOptions` in Nemo .yml config files).
            QFit (:obj:`nemo.signals.QFit`, optional): Object that handles the filter mismatch
                function, *Q*. If not given, the output catalog will not contain `fixed_y_c` columns,
                only `true_y_c` columns.
            wcs (:obj:`astWCS.WCS`, optional): WCS object corresponding to `y0Noise`, if `y0Noise` is
                as noise map (2d image array). Needed if you want the output catalog to contain RA, dec
                coordinates.
            photFilterLabel (:obj:`str`, optional): Name of the reference filter (as defined in the
                Nemo .yml config file) that is used to define y0~ (`fixed_y_c`) and the filter mismatch 
                function, Q.
            tileName (:obj:`str`, optional): Name of the tile for which the sample will be generated.
            SNRLimit (:obj:`float`, optional): Signal-to-noise detection threshold used for the
                output catalog (corresponding to a cut on `fixed_SNR` in Nemo catalogs). Only applied
                if `applySNRCut` is also True (yes, this can be cleaned up).
            makeNames (:obj:`bool`, optional): If True, add names of the form MOCK CL JHHMM.m+/-DDMM
                to the output catalog.
            z (:obj:`float`, optional): If given produce a sample at the nearest z in the MockSurvey
                z grid. The default behaviour is to use the full redshift grid specified by `self.z`.
            numDraws (:obj:`int`, optional): If given, the number of draws to perform from the mass
                function, divided equally among the redshift bins. The default is to use the values
                contained in `self.numClustersByRedshift`.
            areaDeg2 (:obj:`float`, optional): If given, the cluster counts will be scaled to this
                area. Otherwise, they correspond to `self.areaDeg2`. This parameter will be ignored
                if `numDraws` is also given.
            applySNRCut (:obj:`bool`, optional): If True, cut the output catalog according to the
                `fixed_SNR` threshold set by `SNRLimit`.
            applyPoissonScatter (:obj:`bool`, optional): If True, add Poisson noise to the cluster
                counts (implemented by modifiying the number of draws from the mass function).
            applyIntrinsicScatter (:obj:`bool`, optional): If True, apply intrinsic scatter to the
                SZ measurements (`fixed_y_c`), as set by the `sigma_int` parameter in 
                `scalingRelationDict`.
            applyNoiseScatter (:obj:`bool`, optional): If True, apply measurement noise, generated
                from the given noise level or noise map (`y0Noise`), to the output SZ measurements
                (`fixed_y_c`).
            applyRelativisticCorrection (:obj:`bool`, optional): If True, apply the relativistic
                correction.
                
        Returns:
            A catalog as an :obj:`astropy.table.Table` object, in the same format as produced by
            the main `nemo` script.
        
        Notes:
            If both `applyIntrinsicScatter`, `applyNoiseScatter` are set to False, then the output
            catalog `fixed_y_c` values will be exactly the same as `true_y_c`, although each object
            will still have an error bar listed in the output catalog, corresponding to its location
            in the noise map (if given).
                
        """

        t0=time.time()
        if z is None:
            zRange=self.z
        else:
            # Pick the nearest z on the grid
            zIndex=np.argmin(abs(z-self.z))
            zRange=[self.z[zIndex]]

        # Add Poisson noise
        if applyPoissonScatter == True:
            numClusters=np.random.poisson(lam = self.numClusters)
        else:
            numClusters=int(round(self.numClusters))

        if areaDeg2 is not None:
            numClusters=int(round(numClusters*(areaDeg2/self.areaDeg2)))

        if numDraws is not None:
            numClusters=numDraws

        tenToA0, B0, Mpivot, sigma_int=[scalingRelationDict['tenToA0'], scalingRelationDict['B0'],
                                        scalingRelationDict['Mpivot'], scalingRelationDict['sigma_int']]

        # If given y0Noise as RMSMap, draw coords (assuming clusters aren't clustered - which they are...)
        # NOTE: switched to using valid part of RMSMap here rather than areaMask - we need to fix the latter to same area
        # It isn't a significant issue though
        if type(y0Noise) == np.ndarray and y0Noise.ndim == 2:
            # This generates even density RA, dec coords on the whole sky taking into account the projection
            # Consequently, this is inefficient if fed individual tiles rather than a full sky noise map
            assert(wcs is not None)
            RMSMap=y0Noise
            xsList=[]
            ysList=[]
            maxCount=10000
            count=0
            while(len(xsList) < numClusters):
                count=count+1
                if count > maxCount:
                    raise Exception("Failed to generate enough random coords in %d iterations" % (maxCount))
                theta=np.degrees(np.pi*2*np.random.uniform(0, 1, numClusters))
                phi=np.degrees(np.arccos(2*np.random.uniform(0, 1, numClusters)-1))-90
                xyCoords=np.array(wcs.wcs2pix(theta, phi))
                xs=np.array(np.round(xyCoords[:, 0]), dtype = int)
                ys=np.array(np.round(xyCoords[:, 1]), dtype = int)
                mask=np.logical_and(np.logical_and(xs >= 0, xs < RMSMap.shape[1]), np.logical_and(ys >= 0, ys < RMSMap.shape[0]))
                xs=xs[mask]
                ys=ys[mask]
                mask=RMSMap[ys, xs] > 0
                xsList=xsList+xs[mask].tolist()
                ysList=ysList+ys[mask].tolist()
            xs=np.array(xsList)[:numClusters]
            ys=np.array(ysList)[:numClusters]
            del xsList, ysList
            RADecCoords=wcs.pix2wcs(xs, ys)
            RADecCoords=np.array(RADecCoords)
            RAs=RADecCoords[:, 0]
            decs=RADecCoords[:, 1]
            y0Noise=RMSMap[ys, xs]
        elif type(y0Noise) == atpy.Table:
            noisetck=interpolate.splrep(np.cumsum(y0Noise['areaDeg2']/y0Noise['areaDeg2'].sum()), y0Noise['y0RMS'], k = 1)
            rnd=np.random.uniform(0, 1, numClusters)
            vals=interpolate.splev(rnd, noisetck, ext = 3)
            if np.any(vals < 0) or np.any(vals == np.nan):
                raise Exception("Failed to make interpolating spline for RMSTab in tileName = %s" % (tileName))
            y0Noise=vals
            RAs=np.zeros(numClusters)
            decs=np.zeros(numClusters)
        else:
            y0Noise=np.ones(numClusters)*y0Noise
            RAs=np.zeros(numClusters)
            decs=np.zeros(numClusters)

        # Fancy names or not?
        if makeNames == True:
            names=[]
            for RADeg, decDeg in zip(RAs, decs):
                names.append(catalogs.makeName(RADeg, decDeg, prefix = 'MOCK-CL'))
        else:
            names=np.arange(numClusters)+1

        # Draw zs and masses using rejection sampling (from Inigo Zubeldia)
        # NOTE: We can't parallelize this with e.g. multiprocessing because we can't pickle swig objects from pyccl
        zs=np.zeros(y0Noise.shape)
        zErrs=np.zeros(y0Noise.shape)
        log10Ms=np.zeros(y0Noise.shape)
        log10M500cs=np.zeros(y0Noise.shape)
        Qs=np.zeros(y0Noise.shape)
        fRels=np.zeros(y0Noise.shape)
        t0=time.time()
        for clusterIndex in range(numClusters):
            log10Ms[clusterIndex], zs[clusterIndex], log10M500cs[clusterIndex], Qs[clusterIndex], fRels[clusterIndex]=self._drawSampleRow(QFit = QFit)
        t1=time.time()

        # For some cosmo parameters, fRel can wander outside its range for crazy masses
        # So we just cap it at 0.1 here just to avoid -ve in log
        fRels[fRels <= 0]=0.1
        fRels[fRels > 1]=1.0
        Ez2s=np.power(ccl.h_over_h0(self.cosmoModel, 1/(1+zs)), 2)
        try:
            true_y0s=tenToA0*Ez2s*np.power(np.power(10, log10Ms)/Mpivot, 1+B0)*Qs
        except:
            raise Exception("Negative y0 values (probably spline related) for H0 = %.6f Om0 = %.6f sigma8 = %.6f at z = %.3f" % (self.H0, self.Om0, self.sigma8, zk))
        if applyRelativisticCorrection == True:
            true_y0s=true_y0s*fRels

        # Add noise and intrinsic scatter everywhere
        if applyIntrinsicScatter == True:
            scattered_y0s=np.exp(np.random.normal(np.log(true_y0s), sigma_int, len(true_y0s)))
        else:
            scattered_y0s=true_y0s
        if applyNoiseScatter == True:
            measured_y0s=np.random.normal(scattered_y0s, y0Noise)
        else:
            measured_y0s=scattered_y0s

        # NOTE: We're now allowing user to specify mass definition rather than hardcoding M500c
        # So, label the output true mass column appropriately
        massColLabel="true_M%d%s" % (self.delta, self.rhoType[0])
        tab=atpy.Table()
        tab.add_column(atpy.Column(names, 'name'))
        tab.add_column(atpy.Column(RAs, 'RADeg'))
        tab.add_column(atpy.Column(decs, 'decDeg'))
        tab.add_column(atpy.Column(np.power(10, log10Ms)/1e14, massColLabel))
        if 'true_M500c' not in tab.keys():
            tab.add_column(atpy.Column(np.power(10, log10M500cs)/1e14, 'true_M500c'))
        if QFit is None:
            tab.add_column(atpy.Column(true_y0s/1e-4, 'true_y_c'))
        else:
            tab.add_column(atpy.Column(Qs, 'true_Q'))
            tab.add_column(atpy.Column(true_y0s/1e-4, 'true_fixed_y_c'))
            tab.add_column(atpy.Column(measured_y0s/1e-4, 'fixed_y_c'))
            tab.add_column(atpy.Column(y0Noise/1e-4, 'fixed_err_y_c'))
            tab['true_fixed_SNR']=tab['true_fixed_y_c']/tab['fixed_err_y_c']  # True truth, but pre-intrinsic and measurement scatter
            # tab['true_fixed_SNR']=(scattered_y0s/1e-4)/tab['fixed_err_y_c']     # With intrinsic scatter, no measurement scatter
            # tab['true_fixed_SNR']=tab['fixed_y_c']/tab['fixed_err_y_c']         # Like forced photometry case on a real map at true location
            # Apply optimization bias first, then it'll feed through to SNR automatically
            if biasModel is not None:
                corrFactors=biasModel['func'](tab['true_fixed_SNR'], biasModel['params'][0], biasModel['params'][1], biasModel['params'][2])
                tab['fixed_y_c']=tab['fixed_y_c']*corrFactors
            tab['fixed_SNR']=tab['fixed_y_c']/tab['fixed_err_y_c']

        tab.add_column(atpy.Column(zs, 'redshift'))
        tab.add_column(atpy.Column(zErrs, 'redshiftErr'))
        if photFilterLabel is not None and tileName is not None:
            tab.add_column(atpy.Column([photFilterLabel]*len(tab), 'template'))
            tab.add_column(atpy.Column([tileName]*len(tab), 'tileName'))

        # Apply selection?
        if applySNRCut == True:
            selMask=tab['fixed_SNR'] > tab['fixed_err_y_c']*SNRLimit
            tab=tab[selMask]
        t1=time.time()

        return tab

    def _drawSampleRow(self, QFit = None):
        hmfEval=0.
        hmfSample=1.
        while hmfSample > hmfEval:
            ln10MSample=np.random.rand()*(self.log10M[-1]-self.log10M[0])+self.log10M[0]
            zSample=np.random.rand()*(self.z[-1]-self.z[0])+self.z[0]
            hmfSample=np.random.rand()*(self.HMFRange[-1]-self.HMFRange[0])+self.HMFRange[0]
            if self.theoryCode == 'CCL':
                aSample=1/(1+zSample)
                norm_mfunc=1. / np.log(10)
                dVdzdOmega=(((signals.CLight/1000)/self.cosmoModel['H0'])*np.power(1+zSample, 2)*self.cosmoModel.angular_diameter_distance(aSample)**2)/self.cosmoModel.h_over_h0(aSample)
                dndlnM=self.mfunc(self.cosmoModel, np.power(10, ln10MSample), aSample) * norm_mfunc
                hmfEval=4.*np.pi*self.fsky*dVdzdOmega*dndlnM
            elif self.theoryCode == 'CLASS-SZ':
                lnMhSample=np.log(10**ln10MSample*self.cosmoCLASS.h())
                hmfEval=4.*np.pi*self.fsky*self.cosmoCLASS.get_volume_dVdzdOmega_at_z(zSample)*self.cosmoCLASS.get_dndlnM_at_z_and_M(zSample,np.exp(lnMhSample))
        log10M=ln10MSample
        z=zSample
        # Add other quantities - where these are interpolated, they are on a z grid
        zIndex=np.argmin(abs(z-self.z))
        zk=self.z[zIndex]
        if self.delta == 500 and self.rhoType == "critical":
            log10M500c=log10M
        else:
            log10M500c=np.log10(self._transToM500c(self.cosmoModel, np.power(10, log10M), 1/(1+zk)))
        theta500=interpolate.splev(log10M500c, self.theta500Splines[zIndex], ext = 3)
        if QFit is not None:
            Q=QFit.getQ(theta500, z = zk, tileName = tileName)
        else:
            Q=1.0
        fRel=interpolate.splev(log10M500c, self.fRelSplines[zIndex], ext = 3)
        return log10M, z, log10M500c, Q, fRel



"""

An example of how to use the nemo selection function

"""

import os
import sys
import numpy as np
import pylab as plt
import astropy.table as atpy
from astLib import *
from scipy import stats
from scipy import interpolate
from nemo import completeness
import argparse
import time
import IPython

#------------------------------------------------------------------------------------------------------------
def printNumClusters(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = None):
    """Example of how to update selection function.
    
    """
    
    # This function has to be called every time a parameter value is changed
    selFn.update(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = scalingRelationDict)
    
    print(">>> H0 = %.2f km/s/Mpc Om0 = %.2f Ob0 = %.2f sigma_8 = %.2f tenToA0 = %.3e sigma_int = %.2f" \
        % (selFn.mockSurvey.H0, selFn.mockSurvey.Om0, selFn.mockSurvey.Ob0, selFn.mockSurvey.sigma8, 
           selFn.scalingRelationDict['tenToA0'], selFn.scalingRelationDict['sigma_int']))
    numClusters=selFn.mockSurvey.calcNumClustersExpected(compMz = selFn.compMz)
    print("... number of clusters expected = %.2f (in %.2f square degrees) ..." %  (numClusters, selFn.totalAreaDeg2))
    
#------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':

    parser=argparse.ArgumentParser("selFnExample.py")
    parser.add_argument("selFnDir", help="""Directory containing files needed for computing the selection 
                        function.""")
    parser.add_argument("-c", "--config", dest = "configFileName", help="""A .yml configuration file. If
                        this is not given, the config.yml file in selFnDir will be used.""",
                        default = None)
    parser.add_argument("-f", "--footprint", dest = "footprint", help="""Footprint to use, e.g., DES,
                        HSC, KiDS (default: full).""", default = None)
    parser.add_argument("-S", "--SNR-cut", dest = "SNRCut", help="""Use only clusters with fixed_SNR > 
                        this value.""", default = 5.0, type = float)
    parser.add_argument("-z", "--z-step", dest = "zStep", help="""Redshift bin width (default: 0.1).""", 
                        default = 0.02, type = float)
    args = parser.parse_args()
    
    configFileName=args.configFileName
    selFnDir=args.selFnDir
    SNRCut=args.SNRCut
    zStep=args.zStep
    footprintLabel=args.footprint
    
    print(">>> Setting up SNR > %.2f selection function ..." % (SNRCut))
    selFn=completeness.SelFn(selFnDir, SNRCut, configFileName = configFileName, 
                             footprint = footprintLabel, zStep = zStep)
    massLabel="M%d%s" % (selFn.mockSurvey.delta, selFn.mockSurvey.rhoType[0])

    # If we want to play with scaling relation also
    scalingRelationDict=selFn.scalingRelationDict
    
    # Default parameters    
    t0=time.time()
    H0, Om0, Ob0, sigma_8, n_s = 70.0, 0.30, 0.05, 0.80, 0.96
    printNumClusters(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = scalingRelationDict)
    t1=time.time()
    print("... iteration took %.3f sec ..." % (t1-t0))
    completeness.makeMzCompletenessPlot(selFn.compMz, selFn.mockSurvey.log10M, selFn.mockSurvey.z, "test0", massLabel, "test0_Mz.pdf")
    
    # A complete iteration of changing cosmological parameters    
    t0=time.time()
    H0, Om0, Ob0, sigma_8, n_s = 72.0, 0.27, 0.05, 0.74, 1.00
    printNumClusters(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = scalingRelationDict)
    t1=time.time()
    print("... iteration took %.3f sec ..." % (t1-t0))
    completeness.makeMzCompletenessPlot(selFn.compMz, selFn.mockSurvey.log10M, selFn.mockSurvey.z, "test1", massLabel, "test1_Mz.pdf")
    
    # Changing the scaling relation normalisation
    t0=time.time()
    scalingRelationDict['tenToA0']=3e-5
    printNumClusters(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = scalingRelationDict)
    t1=time.time()
    print("... iteration took %.3f sec ..." % (t1-t0))
    completeness.makeMzCompletenessPlot(selFn.compMz, selFn.mockSurvey.log10M, selFn.mockSurvey.z, "test2", massLabel, "test2_Mz.pdf")

    # Changing the scaling relation intrinsic scatter
    t0=time.time()
    scalingRelationDict['tenToA0']=4.95e-5
    scalingRelationDict['sigma_int']=0.001
    printNumClusters(H0, Om0, Ob0, sigma_8, n_s, scalingRelationDict = scalingRelationDict)
    t1=time.time()
    print("... iteration took %.3f sec ..." % (t1-t0))
    completeness.makeMzCompletenessPlot(selFn.compMz, selFn.mockSurvey.log10M, selFn.mockSurvey.z, "test3", massLabel, "test3_Mz.pdf")
    



"""

Makes redshifts.fits catalog (table of names, redshifts matched against nemo output catalog), to allow 
nemoMass to run.

"""

import os
import sys
import argparse
import astropy.table as atpy
from astropy.coordinates import SkyCoord
from astropy.coordinates import match_coordinates_sky
import numpy as np
import IPython

parser=argparse.ArgumentParser()
parser.add_argument("nemoCatalog")
parser.add_argument("redshiftCatalog")
parser.add_argument("-O", "--output-catalog", type=str, default="redshifts.fits")
args=parser.parse_args()

xMatchRadiusArcmin=1.0  # Conservative

tab=atpy.Table().read(args.nemoCatalog)
xTab=atpy.Table().read(args.redshiftCatalog)
cat1=SkyCoord(ra = tab['RADeg'], dec = tab['decDeg'], unit = 'deg')
xMatchRadiusDeg=xMatchRadiusArcmin/60.
cat2=SkyCoord(ra = xTab['RADeg'].data, dec = xTab['decDeg'].data, unit = 'deg')
xIndices, rDeg, sep3d = match_coordinates_sky(cat1, cat2, nthneighbor = 1)
matchTab=xTab[xIndices]

outTab=atpy.Table()
outTab.add_column(tab['name'])
outTab.add_column(tab['RADeg'])
outTab.add_column(tab['decDeg'])
outTab.add_column(atpy.Column(matchTab['z'].data, 'redshift'))
if 'zErr' in matchTab.keys():
    outTab.add_column(atpy.Column(matchTab['zErr'].data, 'redshiftErr'))
else:
    outTab.add_column(atpy.Column(np.zeros(len(tab)), 'redshiftErr'))

mask=np.less(rDeg.value, xMatchRadiusDeg)
outTab=outTab[mask]

outTab.write(args.output_catalog, overwrite = True)
